#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
proto.btc.regtest: Coin daemon regression test mode setup and operations
"""

import os, shutil, json
from ...util import msg, gmsg, die, capfirst, suf
from ...util2 import cliargs_convert
from ...protocol import init_proto
from ...rpc import rpc_init
from ...rpc.util import json_encoder
from ...objmethods import MMGenObject
from ...daemon import CoinDaemon

def create_data_dir(cfg, data_dir):
	try:
		os.stat(os.path.join(data_dir, 'regtest'))
	except:
		pass
	else:
		from ...ui import keypress_confirm
		if keypress_confirm(
				cfg,
				f'Delete your existing MMGen regtest setup at {data_dir!r} and create a new one?'):
			shutil.rmtree(data_dir)
		else:
			die(1, 'Exiting')

	try:
		os.makedirs(data_dir)
	except:
		pass

class MMGenRegtest(MMGenObject):

	rpc_user     = 'bobandalice'
	rpc_password = 'hodltothemoon'
	users        = ('bob', 'alice', 'carol', 'miner')
	coins        = ('btc', 'bch', 'ltc')
	usr_cmds     = (
		'setup',
		'generate',
		'send',
		'start',
		'stop',
		'state',
		'balances',
		'mempool',
		'cli',
		'wallet_cli')
	bdb_hdseed = 'beadcafe' * 8
	bdb_miner_wif = 'cTyMdQ2BgfAsjopRVZrj7AoEGp97pKfrC2NkqLuwHr4KHfPNAKwp'
	bdb_miner_addrs = {
		# cTyMdQ2BgfAsjopRVZrj7AoEGp97pKfrC2NkqLuwHr4KHfPNAKwp hdseed=1
		'btc': 'bcrt1qaq8t3pakcftpk095tnqfv5cmmczysls024atnd',
		'ltc': 'rltc1qaq8t3pakcftpk095tnqfv5cmmczysls05c8zyn',
		'bch': 'n2fxhNx27GhHAWQhyuZ5REcBNrJqCJsJ12',
	}

	def __init__(self, cfg, coin, *, bdb_wallet=False):
		self.cfg = cfg
		self.coin = coin.lower()
		self.bdb_wallet = bdb_wallet

		assert self.coin in self.coins, f'{coin!r}: invalid coin for regtest'

		self.proto = init_proto(cfg, self.coin, regtest=True, need_amt=True)
		self.d = CoinDaemon(
			cfg,
			network_id = self.coin + '_rt',
			test_suite = cfg.test_suite,
			opts       = ['bdb_wallet'] if bdb_wallet else None)

	# Caching creates problems (broken pipe) when recreating + loading wallets,
	# so reinstantiate with every call:
	@property
	async def rpc(self):
		return await rpc_init(self.cfg, self.proto, backend=None, daemon=self.d)

	@property
	async def miner_addr(self):
		if not hasattr(self, '_miner_addr'):
			self._miner_addr = (
				self.bdb_miner_addrs[self.coin] if self.bdb_wallet else
				await self.rpc_call('getnewaddress', wallet='miner'))
		return self._miner_addr

	@property
	async def miner_wif(self):
		if not hasattr(self, '_miner_wif'):
			self._miner_wif = (
				self.bdb_miner_wif if self.bdb_wallet else
				await self.rpc_call('dumpprivkey', (await self.miner_addr), wallet='miner'))
		return self._miner_wif

	def create_hdseed_wif(self):
		from ...tool.api import tool_api
		t = tool_api(self.cfg)
		t.init_coin(self.proto.coin, self.proto.network)
		t.addrtype = 'compressed' if self.proto.coin == 'BCH' else 'bech32'
		return t.hex2wif(self.bdb_hdseed)

	async def generate(self, blocks=1, *, silent=False):

		blocks = int(blocks)

		if self.d.state == 'stopped':
			die(1, 'Regtest daemon is not running')

		self.d.wait_for_state('ready')

		# very slow with descriptor wallet and large block count - 'generatetodescriptor' no better
		out = await self.rpc_call(
			'generatetoaddress',
			blocks,
			await self.miner_addr,
			wallet = 'miner')

		if len(out) != blocks:
			die(4, 'Error generating blocks')

		if not silent:
			gmsg(f'Mined {blocks} block{suf(blocks)}')

	async def create_wallet(self, user):
		return await (await self.rpc).icall(
			'createwallet',
			wallet_name     = user,
			blank           = user != 'miner' or self.bdb_wallet,
			no_keys         = user != 'miner',
			descriptors     = not self.bdb_wallet,
			load_on_startup = False)

	async def setup(self):

		try:
			os.makedirs(self.d.datadir)
		except:
			pass

		if self.d.state != 'stopped':
			await self.rpc_call('stop')

		create_data_dir(self.cfg, self.d.datadir)

		gmsg(f'Starting {self.coin.upper()} regtest setup')

		self.d.start(silent=True)

		for user in ('miner', 'bob', 'alice'):
			gmsg(f'Creating {capfirst(user)}’s tracking wallet')
			await self.create_wallet(user)

		# BCH and LTC daemons refuse to set HD seed with empty blockchain ("in IBD" error),
		# so generate a block:
		await self.generate(1)

		# Unfortunately, we don’t get deterministic output with BCH and LTC even with fixed
		# hdseed, as their 'sendtoaddress' calls produce non-deterministic TXIDs due to random
		# input ordering and fee estimation.
		if self.bdb_wallet:
			await (await self.rpc).call(
				'sethdseed',
				True,
				self.create_hdseed_wif(),
				wallet = 'miner')

		# Broken litecoind can only mine 431 blocks in regtest mode, so generate just enough
		# blocks to fund the test suite.  Generation is slow, so divide into chunks:
		for n in (100, 100, 100, 92): # 392 blocks
			await self.generate(n)

		gmsg('Setup complete')

		if self.cfg.setup_no_stop_daemon:
			msg('Leaving regtest daemon running')
		else:
			msg('Stopping regtest daemon')
			await self.rpc_call('stop')

	def init_daemon(self, *, reindex=False):
		if reindex:
			self.d.usr_coind_args.append('--reindex')

	async def start_daemon(self, *, reindex=False, silent=True):
		self.init_daemon(reindex=reindex)
		self.d.start(silent=silent)
		for user in ('miner', 'bob', 'alice'):
			msg(f'Loading {capfirst(user)}’s wallet')
			await self.rpc_call('loadwallet', user, start_daemon=False)

	async def rpc_call(self, *args, wallet=None, start_daemon=True):
		if start_daemon and self.d.state == 'stopped':
			await self.start_daemon()
		return await (await self.rpc).call(*args, wallet=wallet)

	async def start(self):
		if self.d.state == 'stopped':
			await self.start_daemon(silent=False)
		else:
			msg(f'{self.cfg.coin} regtest daemon already started')

	async def stop(self):
		if self.d.state == 'stopped':
			msg(f'{self.cfg.coin} regtest daemon already stopped')
		else:
			msg(f'Stopping {self.cfg.coin} regtest daemon')
			self.d.stop(silent=True)

	def state(self):
		msg(self.d.state)

	async def balances(self):
		bal = {}
		users = ('bob', 'alice')
		for user in users:
			out = await self.rpc_call('listunspent', 0, wallet=user)
			bal[user] = sum(self.proto.coin_amt(e['amount']) for e in out)

		fs = '{:<16} {:18.8f}'
		for user in users:
			msg(fs.format(user.capitalize()+"'s balance:", bal[user]))
		msg(fs.format('Total balance:', sum(v for k, v in bal.items())))

	async def send(self, addr, amt):
		gmsg(f'Sending {amt} miner {self.d.coin} to address {addr}')
		await self.rpc_call('sendtoaddress', addr, str(amt), wallet='miner')
		await self.generate(1)

	async def mempool(self):
		await self.cli('getrawmempool')

	async def cli(self, *args):
		ret = await self.rpc_call(*cliargs_convert(args))
		print(ret if isinstance(ret, str) else json.dumps(ret, cls=json_encoder, indent=4))

	async def wallet_cli(self, wallet, *args):
		ret = await self.rpc_call(*cliargs_convert(args), wallet=wallet)
		print(ret if isinstance(ret, str) else json.dumps(ret, cls=json_encoder, indent=4))

	async def cmd(self, args):
		ret = getattr(self, args[0])(*args[1:])
		return (await ret) if type(ret).__name__ == 'coroutine' else ret

	async def fork(self, coin): # currently disabled

		proto = init_proto(self.cfg, coin, testnet=False)
		if not [f for f in proto.forks if f[2] == proto.coin.lower() and f[3] is True]:
			die(1, f'Coin {proto.coin} is not a replayable fork of coin {coin}')

		gmsg(f'Creating fork from coin {coin} to coin {proto.coin}')

		source_rt = MMGenRegtest(self.cfg, coin, bdb_wallet=self.bdb_wallet)

		try:
			os.stat(source_rt.d.datadir)
		except:
			die(1, f'Source directory {source_rt.d.datadir!r} does not exist!')

		# stop the source daemon
		if source_rt.d.state != 'stopped':
			await source_rt.d.cli('stop')

		# stop our daemon
		if self.d.state != 'stopped':
			await self.rpc_call('stop')

		try:
			os.makedirs(self.d.datadir)
		except:
			pass

		create_data_dir(self.cfg, self.d.datadir)
		os.rmdir(self.d.datadir)
		shutil.copytree(source_rt.d.datadir, self.d.datadir, symlinks=True)
		await self.start_daemon(reindex=True)
		await self.rpc_call('stop')

		gmsg(f'Fork {proto.coin} successfully created')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txsign: Sign a transaction generated by 'mmgen-txcreate'
"""

from .cfg import gc, Config
from .util import msg, ymsg, suf, die, async_run
from .subseed import SubSeedIdxRange
from .color import orange

opts_data = {
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc':    f'Sign cryptocoin transactions generated by {gc.proj_name.lower()}-txcreate',
		'usage':   '[opts] <transaction file>... [seed source]...',
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long (global) options
-a, --autosign        Sign a transaction created for offline autosigning (see
                      ‘mmgen-autosign’)
-b, --brain-params=l,p Use seed length 'l' and hash preset 'p' for
                      brainwallet input
-d, --outdir=      d  Specify an alternate directory 'd' for output
-D, --tx-id           Display transaction ID and exit
-e, --echo-passphrase Print passphrase to screen when typing it
-E, --use-internal-keccak-module Force use of the internal keccak module
-i, --in-fmt=      f  Input is from wallet format 'f' (see FMT CODES below)
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated)
-O, --old-incog-fmt   Specify old-format incognito input
-l, --seed-len=    l  Specify wallet seed length of 'l' bits. This option
                      is required only for brainwallet and incognito inputs
                      with non-standard (< {dsl}-bit) seed lengths.
-p, --hash-preset=p   Use the scrypt hash parameters defined by preset 'p'
                      for password hashing (default: '{gc.dfl_hash_preset}')
-z, --show-hash-presets Show information on available hash presets
-k, --keys-from-file=f Provide additional keys for non-{pnm} addresses
-K, --keygen-backend=n Use backend 'n' for public key generation.  Options
                      for {coin_id}: {kgs}
-M, --mmgen-keys-from-file=f Provide keys for {pnm} addresses in a key-
                      address file (output of '{pnl}-keygen'). Permits
                      online signing without an {pnm} seed source. The
                      key-address file is also used to verify {pnm}-to-{cu}
                      mappings, so the user should record its checksum.
-P, --passwd-file= f  Get {pnm} wallet passphrase from file 'f'
-q, --quiet           Suppress warnings; overwrite files without prompting
-I, --info            Display information about the transaction and exit
-t, --terse-info      Like '--info', but produce more concise output
-u, --subseeds=     n The number of subseed pairs to scan for (default: {ss},
                      maximum: {ss_max}). Only the default or first supplied
                      wallet is scanned for subseeds.
-v, --verbose         Produce more verbose output
-V, --vsize-adj=   f  Adjust transaction's estimated vsize by factor 'f'
-W, --allow-non-wallet-swap Allow signing of swap transactions that send funds
                      to non-wallet addresses
-y, --yes             Answer 'yes' to prompts, suppress non-essential output
""",
	'notes': """
{t}
Seed source files must have the canonical extensions listed in the 'FileExt'
column below:

{f}
"""
	},
	'code': {
		'options': lambda cfg, proto, help_notes, s: s.format(
			cfg     = cfg,
			gc      = gc,
			pnm     = gc.proj_name,
			pnl     = gc.proj_name.lower(),
			kgs     = help_notes('keygen_backends'),
			coin_id = help_notes('coin_id'),
			dsl     = help_notes('dfl_seed_len'),
			ss      = help_notes('dfl_subseeds'),
			ss_max  = SubSeedIdxRange.max_idx,
			cu      = proto.coin),
		'notes': lambda help_mod, help_notes, s: s.format(
			t       = help_mod('txsign'),
			f       = help_notes('fmt_codes')),
	}
}

cfg = Config(opts_data=opts_data)

if not cfg._args:
	cfg._usage()

from .fileutil import check_infile
for i in cfg._args:
	check_infile(i)

if not cfg.info and not cfg.terse_info:
	from .ui import do_license_msg
	do_license_msg(cfg, immed=True)

from .tx.keys import TxKeys, pop_txfiles, pop_seedfiles

txfiles = pop_txfiles(cfg)
seedfiles = pop_seedfiles(cfg)

async def main():

	bad_tx_count = 0
	tx_num_disp = ''

	for tx_num, tx_file in enumerate(txfiles, 1):

		if len(txfiles) > 1:
			tx_num_disp = f' #{tx_num}'
			msg(orange(f'\nTransaction{tx_num_disp} of {len(txfiles)}:'))

		from .tx import UnsignedTX
		tx1 = UnsignedTX(cfg=cfg, filename=tx_file)

		cfg._util.vmsg(f'Successfully opened transaction file {tx_file!r}')

		if tx1.proto.sign_mode == 'daemon':
			from .rpc import rpc_init
			tx1.rpc = await rpc_init(cfg, tx1.proto, ignore_wallet=True)

		if cfg.tx_id:
			msg(tx1.txid)
			continue

		if cfg.info or cfg.terse_info:
			tx1.view(pause=False, terse=cfg.terse_info)
			continue

		if not cfg.yes:
			tx1.info.view_with_prompt(f'View data for transaction{tx_num_disp}?')

		if tx2 := await tx1.sign(TxKeys(cfg, tx1, seedfiles=seedfiles).keys, tx_num_disp):
			if not cfg.yes:
				tx2.add_comment() # edits an existing comment
			tx2.file.write(ask_write=not cfg.yes, ask_write_default_yes=True, add_desc=tx_num_disp)
		else:
			ymsg('Transaction could not be signed')
			bad_tx_count += 1

	if bad_tx_count:
		die(2, f'{bad_tx_count} transaction{suf(bad_tx_count)} could not be signed')

async_run(main())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
examples/bip_hd.py: Usage examples for the MMGen BIP-32/-44 hierarchical/deterministic library
"""

from mmgen.cfg import Config
from mmgen.util import fmt
from mmgen.bip39 import bip39
from mmgen.bip_hd import MasterNode, BipHDNode

cfg = Config()

bip39_mnemonic = 'cat swing flag economy stadium alone churn speed unique patch report train'

seed = bip39().generate_seed(bip39_mnemonic.split())

m = MasterNode(cfg, seed)

# Derive sample path:

# to_chain() derives default chain for coin/addr_type pair:
dfl_pub_chain = m.to_chain(idx=0, coin='ltc', addr_type='bech32')
dfl_chg_chain = m.to_chain(idx=1, coin='ltc', addr_type='bech32')

print('Default path (LTC, bech32):\n')
print(f'  public chain xpub:\n    {dfl_pub_chain.xpub}\n')
print(f'  internal chain xpub:\n    {dfl_chg_chain.xpub}\n')
print(f'  public chain addr 0:\n    {dfl_pub_chain.derive_public(0).address}\n')
print(f'  public chain addr 1:\n    {dfl_pub_chain.derive_public(1).address}\n')

# Derive sample path using path string:

dfl_pub_chain_from_path = BipHDNode.from_path(
	base_cfg  = cfg,
	seed      = seed,
	# purpose=84 (bech32 [BIP-84]), coin_type=2 (LTC mainnet [SLIP-44]), account=0, chain=0 (public)
	# as per BIP-44, ‘purpose’, ‘coin_type’ and ‘account’ are hardened, while ‘chain’ is not
	path_str  = "m/84'/2'/0'/0",
	coin      = 'ltc',
	addr_type = 'bech32')

assert dfl_pub_chain_from_path.xpub == dfl_pub_chain.xpub

# Derive sample path step-by-step:

# Configure master node with coin/addr_type pair:
master = m.init_cfg(coin='ltc', addr_type='bech32')

# ‘idx’ and ‘hardened’ args may be omitted at depths where defaults exist:
purpose = master.derive_private()     # ‘idx’ is auto-computed from addr_type (BIP-44/49/84)
coin_type = purpose.derive_private()  # ‘idx’ is auto-computed from coin/network (SLIP-44)
account = coin_type.derive_private(idx=0)
pub_chain = account.derive_public(idx=0)

assert pub_chain.xpub == dfl_pub_chain.xpub

# Initialize node from xpub:
pub_chain_from_xpub = BipHDNode.from_extended_key(cfg, 'ltc', pub_chain.xpub)

assert pub_chain_from_xpub.xpub == pub_chain.xpub

# To derive arbitrary BIP-32 paths, ignoring BIP-44, specify ‘no_path_checks’
nonstd_path = BipHDNode.from_path(
	base_cfg  = cfg,
	seed      = seed,
	path_str  = "m/111'/222/333/444",
	coin      = 'eth',
	addr_type = 'E',
	no_path_checks = True)

print('Non-standard path (ETH):\n')
print(f'  xpub:\n    {nonstd_path.xpub}\n')
print(f'  WIF key:\n    {nonstd_path.privkey.wif}\n')
print(f'  address:\n    {nonstd_path.address}\n')

# Display parsed xpub:
parsed_xpub = nonstd_path.key_extended(public=True)
print('Default path parsed xpub:\n')
print(fmt(str(parsed_xpub), indent='  '))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# Project source code repository: https://github.com/mmgen/mmgen-wallet
# Licensed according to the terms of GPL Version 3.  See LICENSE for details.

"""
test.cmdtest_d.input: user input tests for the MMGen cmdtest.py test suite
"""

import sys, os

from mmgen.cfg import gc
from mmgen.util import fmt, capfirst, remove_whitespace
from mmgen.wallet import get_wallet_cls

from ..include.common import (
	imsg,
	imsg_r,
	sample_mn,
	get_data_from_file,
	read_from_file,
	strip_ansi_escapes
)
from .include.common import Ctrl_U, ref_dir
from .base import CmdTestBase
from .include.input import stealth_mnemonic_entry, user_dieroll_entry

hold_protect_delay = 2 if sys.platform == 'darwin' else None

class CmdTestInput(CmdTestBase):
	'user input'
	networks = ('btc',)
	tmpdir_nums = [1]
	color = True
	cmd_group_in = (
		('subgroup.char', []),
		('subgroup.line', []),
		('subgroup.password', []),
		('subgroup.misc', []),
		('subgroup.wallet', []),
		('subgroup.mnemonic', []),
		('subgroup.dieroll', []),
	)
	cmd_subgroups = {
	'char': (
		'get_char() function',
		('get_char1',                    'get_char()'),
		('get_char2',                    'get_char() [multiple characters]'),
		('get_char3',                    'get_char() [no prompt]'),
		('get_char4',                    'get_char() [utf8]'),
		('get_char_term1',               'get_char() [term, utf8]'),
		('get_char_term2',               'get_char() [term, multiple characters]'),
		('get_char_term3',               'get_char() [term, prehold_protect=False]'),
		('get_char_term4',               'get_char() [term, immed_chars="xyz"]'),
	),
	'line': (
		'line_input() function',
		('line_input',                    'line_input()'),
		('line_input_term1',              'line_input() [term]'),
		('line_input_term2',              'line_input() [term, no hold protect]'),
		('line_input_insert',             'line_input() [inserted text]'),
		('line_input_insert_term1',       'line_input() [inserted text, term]'),
		('line_input_insert_term2',       'line_input() [inserted text, term, no hold protect]'),
		('line_input_edit_term',          'line_input() [edited text, term, utf8]'),
		('line_input_edit_term_insert',   'line_input() [inserted + edited text, term, utf8]'),
		('line_input_erase_term',         'line_input() [inserted + erased text, term]'),
	),
	'password': (
		'password entry via line_input()',
		('password_entry_noecho',         'utf8 password entry'),
		('password_entry_noecho_term',    'utf8 password entry [term]'),
		('password_entry_echo',           'utf8 password entry (echoed)'),
		('password_entry_echo_term',      'utf8 password entry (echoed) [term]'),
	),
	'misc': (
		'miscellaneous user-level UI functions',
		('get_seed_from_stdin',           'reading seed phrase from STDIN'),
	),
	'wallet': (
		'hash preset, password and label entry',
		('get_passphrase_ui',             'hash preset, password and label (wallet.py)'),
		('get_passphrase_cmdline',        'hash preset, password and label (wallet.py - from cmdline)'),
		('get_passphrase_crypto',         'hash preset, password and label (crypto.py)'),
	),
	'mnemonic': (
		'mnemonic entry',
		('mnemonic_entry_mmgen',           'stealth mnemonic entry (mmgen)'),
		('mnemonic_entry_mmgen_minimal',   'stealth mnemonic entry (mmgen - minimal entry mode)'),
		('mnemonic_entry_bip39',           'stealth mnemonic entry (bip39)'),
		('mnemonic_entry_bip39_short',     'stealth mnemonic entry (bip39 - short entry mode)'),
		('mn2hex_interactive_mmgen',       'mn2hex_interactive (mmgen)'),
		('mn2hex_interactive_mmgen_fixed', 'mn2hex_interactive (mmgen - fixed (10-letter) entry mode)'),
		('mn2hex_interactive_bip39',       'mn2hex_interactive (bip39)'),
		('mn2hex_interactive_bip39_short', 'mn2hex_interactive (bip39 - short entry mode (+pad entry))'),
		('mn2hex_interactive_bip39_fixed', 'mn2hex_interactive (bip39 - fixed (4-letter) entry mode)'),
		('mn2hex_interactive_xmr',         'mn2hex_interactive (xmrseed)'),
		('mn2hex_interactive_xmr_short',   'mn2hex_interactive (xmrseed - short entry mode)'),
	),
	'dieroll': (
		'dieroll entry',
		('dieroll_entry',                 'dieroll entry (base6d)'),
		('dieroll_entry_usrrand',         'dieroll entry (base6d) with added user entropy'),
	)
	}

	def get_seed_from_stdin(self):
		self.spawn(msg_only=True)
		from subprocess import run, PIPE
		cmd = ['python3', 'cmds/mmgen-walletconv', '--skip-cfg-file', '--in-fmt=words', '--out-fmt=words', '--outdir=test/trash']
		mn = sample_mn['mmgen']['mn']
		run_env = dict(os.environ)
		run_env['MMGEN_TEST_SUITE'] = ''

		cp = run(cmd, input=mn.encode(), stdout=PIPE, stderr=PIPE, env=run_env)

		from mmgen.color import set_vt100
		set_vt100()
		imsg(cp.stderr.decode().strip())
		res = get_data_from_file(self.cfg, 'test/trash/A773B05C[128].mmwords', silent=True).strip()
		assert res == mn, f'{res} != {mn}'
		return 'ok' if b'written to file' in cp.stderr else 'error'

	def get_passphrase_ui(self):
		t = self.spawn('test/misc/get_passphrase.py', ['--usr-randchars=0', 'seed'], cmd_dir='.')

		# 1 - new wallet, default hp, label;empty pw
		t.expect('accept the default.*: ', '\n', regex=True)

		# bad repeat
		t.expect('new MMGen wallet: ', 'pass1\n')
		t.expect('peat passphrase: ', 'pass2\n')

		# good repeat
		t.expect('new MMGen wallet: ', '\n')
		t.expect('peat passphrase: ', '\n')
		t.expect('mpty pass')

		t.expect('no label: ', '\n')

		t.expect('[][3][No Label]')

		# 2 - new wallet, user-selected hp, pw, label
		t.expect('accept the default.*: ', '1\n', regex=True)

		t.expect('new MMGen wallet: ', 'pass1\n')
		t.expect('peat passphrase: ', 'pass1\n')

		t.expect('no label: ', 'lbl1\n')

		t.expect('[pass1][1][lbl1]')

		# 3 - passchg, nothing changes
		t.expect('new hash preset')
		t.expect('reuse the old value.*: ', '\n', regex=True)
		t.expect('unchanged')

		t.expect('new passphrase.*: ', 'pass1\n', regex=True)
		t.expect('peat new passphrase: ', 'pass1\n')
		t.expect('unchanged')

		t.expect('reuse the label .*: ', '\n', regex=True)
		t.expect('unchanged')

		t.expect('[pass1][1][lbl1]')

		# 4 - passchg, everything changes
		t.expect('new hash preset')
		t.expect('reuse the old value.*: ', '2\n', regex=True)
		t.expect(' changed to')

		t.expect('new passphrase.*: ', 'pass2\n', regex=True)
		t.expect('peat new passphrase: ', 'pass2\n')
		t.expect(' changed')

		t.expect('reuse the label .*: ', 'lbl2\n', regex=True)
		t.expect(' changed to')
		t.expect('[pass2][2][lbl2]')

		# 5 - wallet from file
		t.expect('from file')

		# bad passphrase
		t.expect('passphrase for MMGen wallet: ', 'bad\n')
		t.expect('Trying again')

		# good passphrase
		t.expect('passphrase for MMGen wallet: ', 'reference password\n')
		t.expect('[reference password][1][No Label]')

		return t

	def get_passphrase_cmdline(self):
		with open('test/trash/pwfile', 'w') as fp:
			fp.write('reference password\n')
		t = self.spawn('test/misc/get_passphrase.py', [
			'--usr-randchars=0',
			'--label=MyLabel',
			'--passwd-file=test/trash/pwfile',
			'--hash-preset=1',
			'seed'],
			cmd_dir = '.')
		for _ in range(4):
			t.expect('[reference password][1][MyLabel]')

		return t

	def get_passphrase_crypto(self):
		t = self.spawn('test/misc/get_passphrase.py', ['--usr-randchars=0', 'crypto'], cmd_dir='.')

		# new passwd
		t.expect('passphrase for .*: ', 'x\n', regex=True)
		t.expect('peat passphrase: ', '\n')
		t.expect('passphrase for .*: ', 'pass1\n', regex=True)
		t.expect('peat passphrase: ', 'pass1\n')
		t.expect('[pass1]')

		# existing passwd
		t.expect('passphrase for .*: ', 'pass2\n', regex=True)
		t.expect('[pass2]')

		# hash preset
		t.expect('accept the default .*: ', '0\n', regex=True)
		t.expect('nvalid')
		t.expect('accept the default .*: ', '8\n', regex=True)
		t.expect('nvalid')
		t.expect('accept the default .*: ', '7\n', regex=True)
		t.expect('[7]')

		# hash preset (default)
		t.expect('accept the default .*: ', '\n', regex=True)
		t.expect(f'[{gc.dfl_hash_preset}]')

		return t

	def _input_func(self, func_name, arg_dfls, func_args, text, expect, term, delay=None):
		if term and sys.platform == 'win32':
			return ('skip_warn', 'pexpect_spawn not supported on Windows platform')
		func_args = dict(zip(arg_dfls.keys(), func_args))
		t = self.spawn(
			'test/misc/input_func.py',
			[func_name, repr(func_args)],
			cmd_dir='.',
			pexpect_spawn=term)
		imsg('Parameters:')
		imsg(f'  pexpect_spawn: {term}')
		imsg(f'  sending:       {text!r}')
		imsg(f'  expecting:     {expect!r}')
		imsg('\nFunction args:')
		for k, v in func_args.items():
			imsg('  {:14} {!r}'.format(k+':', v))
		imsg_r('\nScript output: ')
		prompt_add = (func_args['insert_txt'] if term else '') if func_name == 'line_input' else ''
		prompt = func_args['prompt'] + prompt_add
		t.expect('Calling ')
		if prompt:
			t.expect(prompt, text, delay=delay)
		else:
			t.send(text, delay=delay)
		ret = t.expect_getend('  ==> ')
		assert ret == repr(expect), f'Text mismatch! {ret} != {repr(expect)}'
		return t

	def _get_char(self, func_args, text, expect, term):
		arg_dfls = {
			'prompt': '',
			'immed_chars': '',
			'prehold_protect': True,
			'num_bytes': 5,
		}
		return self._input_func('get_char', arg_dfls, func_args, text, expect, term)

	def _line_input(self, func_args, text, expect, term, delay=None):
		arg_dfls = {
			'prompt': '', # positional
			'echo': True,
			'insert_txt': '',
			'hold_protect': True,
		}
		return self._input_func('line_input', arg_dfls, func_args, text+'\n', expect, term, delay=delay)

	def get_char1(self):
		return self._get_char(['prompt> ', '', True, 5], 'x', 'x', False)

	def get_char2(self):
		expect = 'x' if sys.platform == 'win32' else 'xxxxx'
		return self._get_char(['prompt> ', '', True, 5], 'xxxxx', expect, False)

	def get_char3(self):
		return self._get_char(['', '', True, 5], 'x', 'x', False)

	def get_char4(self):
		return self._get_char(['prompt> ', '', True, 2], 'α', 'α', False) # UTF-8, must get 2 bytes

	def get_char_term1(self):
		return self._get_char(['prompt> ', '', True, 2], 'β', 'β', True)  # UTF-8, must get 2 bytes

	def get_char_term2(self):
		return self._get_char(['prompt> ', '', True, 5], 'xxxxx', 'xxxxx', True)

	def get_char_term3(self):
		return self._get_char(['', '', False, 5], 'x', 'x', True)

	def get_char_term4(self):
		return self._get_char(['prompt> ', 'xyz', False, 5], 'x', 'x', True)

	def line_input(self):
		return self._line_input(
			['prompt> ', True, '', True],
			'foo',
			'foo',
			False)

	def line_input_term1(self):
		return self._line_input(
			['prompt> ', True, '', True],
			'foo',
			'foo',
			True,
			hold_protect_delay)

	def line_input_term2(self):
		return self._line_input(
			['prompt> ', True, '', False],
			'foo',
			'foo',
			True)

	def line_input_insert(self):
		return self._line_input(
			['prompt> ', True, 'inserted text', True],
			'foo',
			'foo',
			False)

	def line_input_insert_term1(self):
		if self.skip_for_mac('readline text buffer issues'):
			return 'skip'
		return self._line_input(
			['prompt> ', True, 'foo', True],
			'bar',
			'foobar',
			True,
			hold_protect_delay)

	def line_input_insert_term2(self):
		if self.skip_for_mac('readline text buffer issues'):
			return 'skip'
		return self._line_input(
			['prompt> ', True, 'foo', False],
			'bar',
			'foobar',
			True)

	def line_input_edit_term(self):
		return self._line_input(
			['prompt> ', True, '', True],
			'\b\bφυφυ\b\bβαρ',
			'φυβαρ',
			True,
			hold_protect_delay)

	def line_input_edit_term_insert(self):
		if self.skip_for_mac('readline text buffer issues'):
			return 'skip'
		return self._line_input(
			['prompt> ', True, 'φυφυ', True],
			'\b\bβαρ',
			'φυβαρ',
			True,
			hold_protect_delay)

	def line_input_erase_term(self):
		if self.skip_for_mac('readline text buffer issues'):
			return 'skip'
		return self._line_input(
			['prompt> ', True, 'foobarbaz', True],
			Ctrl_U + 'foobar',
			'foobar',
			True,
			hold_protect_delay)

	def _password_entry(self, prompt, opts=[], term=False):
		if term and sys.platform == 'win32':
			return ('skip_warn', 'pexpect_spawn not supported on Windows platform')
		t = self.spawn('test/misc/input_func.py', opts + ['passphrase'], cmd_dir='.', pexpect_spawn=term)
		imsg(f'Terminal: {term}')
		pw = 'abc-α'
		t.expect(prompt, pw+'\n')
		ret = t.expect_getend('Entered: ')
		assert ret == pw, f'Password mismatch! {ret} != {pw}'
		return t

	winskip_msg = """
		pexpect_spawn not supported on Windows platform
		Perform the following test by hand with non-ASCII password abc-α
		or another password in your native alphabet:

		  test/misc/input_func.py{} passphrase
	"""

	def password_entry_noecho(self, term=False):
		return self._password_entry('Enter passphrase: ', term=term)

	def password_entry_noecho_term(self):
		if self.skip_for_win('no pexpect_spawn'):
			return ('skip_warn', '\n' + fmt(self.winskip_msg.format(''), strip_char='\t'))
		return self.password_entry_noecho(term=True)

	def password_entry_echo(self, term=False):
		return self._password_entry('Enter passphrase (echoed): ', ['--echo-passphrase'], term=term)

	def password_entry_echo_term(self):
		if self.skip_for_win('no pexpect_spawn'):
			return ('skip_warn', '\n' + fmt(self.winskip_msg.format(' --echo-passphrase'), strip_char='\t'))
		return self.password_entry_echo(term=True)

	def _mn2hex(self, fmt, entry_mode='full', mn=None, pad_entry=False, enter_for_dfl=False):
		mn = mn or sample_mn[fmt]['mn'].split()
		t = self.spawn('mmgen-tool', ['mn2hex_interactive', 'fmt='+fmt, 'mn_len=12', 'print_mn=1'])
		from mmgen.mn_entry import mn_entry
		mne = mn_entry(self.cfg, fmt, entry_mode=entry_mode)
		t.expect(
			'Type a number.*: ',
			('\n' if enter_for_dfl else str(mne.entry_modes.index(entry_mode)+1)),
			regex = True)
		t.expect(r'Using entry mode (\S+)', regex=True)
		mode = strip_ansi_escapes(t.p.match.group(1)).lower()
		assert mode == mne.em.name.lower(), f'{mode} != {mne.em.name.lower()}'
		stealth_mnemonic_entry(t, mne, mn, entry_mode=entry_mode, pad_entry=pad_entry)
		t.expect(sample_mn[fmt]['hex'])
		return t

	def _user_seed_entry(
			self,
			fmt,
			usr_rand    = False,
			out_fmt     = None,
			entry_mode  = 'full',
			mn          = None,
			seedlen_opt = False):

		wcls = get_wallet_cls(fmt_code=fmt)
		wf = os.path.join(ref_dir, f'FE3C6545.{wcls.ext}')
		if wcls.base_type == 'mnemonic':
			mn = mn or read_from_file(wf).strip().split()
		elif wcls.type == 'dieroll':
			mn = mn or list(remove_whitespace(read_from_file(wf)))
			for idx, val in ((5, 'x'), (18, '0'), (30, '7'), (44, '9')):
				mn.insert(idx, val)
		t = self.spawn(
			'mmgen-walletconv',
			['--usr-randchars=10', '--stdout']
			+ (['--seed-len=128'] if seedlen_opt else [])
			+ [f'--in-fmt={fmt}', f'--out-fmt={out_fmt or fmt}']
		)
		t.expect(f'{capfirst(wcls.base_type or wcls.type)} type:.*{wcls.mn_type}', regex=True)
		if not seedlen_opt:
			t.expect(wcls.choose_seedlen_prompt, '1')
			t.expect('(Y/n): ', 'y')
		if wcls.base_type == 'mnemonic':
			t.expect('Type a number.*: ', '6', regex=True)
			t.expect('invalid')
			from mmgen.mn_entry import mn_entry
			mne = mn_entry(self.cfg, fmt, entry_mode=entry_mode)
			t.expect('Type a number.*: ', str(mne.entry_modes.index(entry_mode)+1), regex=True)
			t.expect(r'Using entry mode (\S+)', regex=True)
			mode = strip_ansi_escapes(t.p.match.group(1)).lower()
			assert mode == mne.em.name.lower(), f'{mode} != {mne.em.name.lower()}'
			stealth_mnemonic_entry(t, mne, mn, entry_mode=entry_mode)
		elif wcls.type == 'dieroll':
			user_dieroll_entry(t, mn)
			if usr_rand:
				t.expect(wcls.user_entropy_prompt, 'y')
				t.usr_rand(10)
			else:
				t.expect(wcls.user_entropy_prompt, 'n')
		if not usr_rand:
			sid_chk = 'FE3C6545'
			sid = t.expect_getend(f'Valid {wcls.desc} for Seed ID')
			sid = strip_ansi_escapes(sid.split(',')[0])
			assert sid_chk in sid, f'Seed ID mismatch! {sid_chk} not found in {sid}'
		t.expect('to confirm: ', 'YES\n')
		return t

	def mnemonic_entry_mmgen_minimal(self):
		from mmgen.mn_entry import mn_entry
		# erase_chars: '\b\x7f'
		m = mn_entry(self.cfg, 'mmgen', entry_mode='minimal')
		np = 2
		mn = (
			'z',
			'aa',
			'1d2ud',
			'fo{}ot{}#'.format('1' * np, '2' * (m.em.pad_max - np)), # substring of 'football'
			'des1p)%erate\n', # substring of 'desperately'
			'#t!(ie',
			'!)sto8o',
			'the123m8!%s',
			'349t(5)rip',
			'di\b\bdesce',
			'cea',
			'bu\x7f\x7fsuic',
			'app\bpl',
			'wd',
			'busy')
		return self._user_seed_entry('words', entry_mode='minimal', mn=mn)
	def mnemonic_entry_mmgen(self):
		return self._user_seed_entry('words', entry_mode='full')
	def mnemonic_entry_bip39(self):
		return self._user_seed_entry('bip39', entry_mode='full')
	def mnemonic_entry_bip39_short(self):
		return self._user_seed_entry('bip39', entry_mode='short')

	def mn2hex_interactive_mmgen(self):
		return self._mn2hex('mmgen', entry_mode='full')
	def mn2hex_interactive_mmgen_fixed(self):
		return self._mn2hex('mmgen', entry_mode='fixed')
	def mn2hex_interactive_bip39(self):
		return self._mn2hex('bip39', entry_mode='full')
	def mn2hex_interactive_bip39_short(self):
		return self._mn2hex('bip39', entry_mode='short', pad_entry=True)
	def mn2hex_interactive_bip39_fixed(self):
		return self._mn2hex('bip39', entry_mode='fixed', enter_for_dfl=True)
	def mn2hex_interactive_xmr(self):
		return self._mn2hex('xmrseed', entry_mode='full')
	def mn2hex_interactive_xmr_short(self):
		return self._mn2hex('xmrseed', entry_mode='short')

	def dieroll_entry(self):
		return self._user_seed_entry('dieroll', seedlen_opt=True)
	def dieroll_entry_usrrand(self):
		return self._user_seed_entry('dieroll', usr_rand=True, out_fmt='bip39')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.unsigned: unsigned transaction class
"""

from .completed import Completed
from ..util import remove_dups

class Unsigned(Completed):
	desc = 'unsigned transaction'
	ext  = 'rawtx'
	automount = False

	def delete_attrs(self, desc, attr):
		for e in getattr(self, desc):
			if hasattr(e, attr):
				delattr(e, attr)

	def get_sids(self, desc):
		return remove_dups(
			(e.mmid.sid for e in getattr(self, desc) if e.mmid),
			quiet = True)

class AutomountUnsigned(Unsigned):
	desc = 'unsigned automount transaction'
	ext  = 'arawtx'
	automount = True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.zec.addrgen: Zcash-Z address generation class for the MMGen suite
"""

from ...addrgen import addr_generator, check_data
from ...addr import CoinAddr
from ..btc.common import b58chk_encode

class zcash_z(addr_generator.base):

	@check_data
	def to_addr(self, data):
		ret = b58chk_encode(
			self.proto.addr_fmt_to_ver_bytes['zcash_z']
			+ data.pubkey)
		return CoinAddr(self.proto, ret)

	@check_data
	def to_viewkey(self, data):
		return self.proto.viewkey(
			b58chk_encode(
				self.proto.addr_fmt_to_ver_bytes['viewkey']
				+ data.viewkey_bytes))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.overlay.__init__: Initialize the MMGen test suite overlay tree
"""

import sys, os, shutil

def get_overlay_tree_dir(repo_root):
	return os.path.join(repo_root, 'test', 'overlay', 'tree')

def overlay_setup(repo_root):

	def process_srcdir(pkgname, d):
		relpath = d.split('.')
		srcdir = os.path.join(repo_root, *relpath)
		destdir = os.path.join(overlay_tree_dir, *relpath)
		fakemod_dir = os.path.join(fakemod_root, pkgname, *(relpath[1:]))
		os.makedirs(destdir)
		for fn in os.listdir(srcdir):
			if (
				fn.endswith('.py') or
				d == f'{pkgname}.data' or
				(d == 'mmgen.proto.secp256k1' and fn.startswith('secp256k1'))
			):
				if fn.endswith('.py') and os.path.exists(os.path.join(fakemod_dir, fn)):
					make_link(
						os.path.join(fakemod_dir, fn),
						os.path.join(destdir, fn))
					link_fn = fn.removesuffix('.py') + '_orig.py'
				else:
					link_fn = fn
				make_link(
					os.path.join(srcdir, fn),
					os.path.join(destdir, link_fn))

	overlay_tree_dir = get_overlay_tree_dir(repo_root)
	fakemod_root = os.path.join(repo_root, 'test', 'overlay', 'fakemods')
	common_path = os.path.join(os.path.sep, 'test', 'overlay', 'fakemods')
	pkgdata = ((
			os.path.realpath(e.path).removesuffix(os.path.join(common_path, e.name)),
			e.name
		) for e in os.scandir(fakemod_root) if e.is_dir())

	for repodir, pkgname in pkgdata:
		if not os.path.exists(os.path.join(overlay_tree_dir, pkgname)):

			sys.stderr.write(f'Setting up overlay tree: {pkgname}\n')

			make_link = os.symlink if sys.platform in ('linux', 'darwin') else shutil.copy2
			shutil.rmtree(os.path.join(overlay_tree_dir, pkgname), ignore_errors=True)

			import configparser
			cfg = configparser.ConfigParser()
			cfg.read(os.path.join(repodir, 'setup.cfg'))

			for d in cfg.get('options', 'packages').split():
				process_srcdir(pkgname, d)

	sys.path.insert(0, overlay_tree_dir)

	return overlay_tree_dir



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-addrgen: Generate a series or range of addresses from an MMGen
               deterministic wallet
"""

from . import addrlist
from .cfg import gc, Config
from .addr import MMGenAddrType
from .wallet import Wallet

if gc.prog_name == 'mmgen-keygen':
	gen_what = 'keys'
	gen_clsname = 'KeyAddrList'
	gen_desc = 'secret keys'
	filter_codes = ['-', 'k']
	note_addrkey = 'By default, both addresses and secret keys are generated.\n\n'
else:
	gen_what = 'addresses'
	gen_clsname = 'AddrList'
	gen_desc = 'addresses'
	filter_codes = ['-']
	note_addrkey = ''

opts_data = {
	'filter_codes': filter_codes,
	'sets': [('print_checksum', True, 'quiet', True)],
	'text': {
		'desc': f"""
                 Generate a range or list of {gen_desc} from an {gc.proj_name} wallet,
                 mnemonic, seed or brainwallet
			  """,
		'usage':'[opts] [seed source] <index list or range(s)>',
		'options': """
			-- -h, --help            Print this help message
			-- --, --longhelp        Print help message for long (global) options
			-k -A, --no-addresses    Print only secret keys, no addresses
			-- -c, --print-checksum  Print address list checksum and exit
			-- -d, --outdir=      d  Output files to directory 'd' instead of working dir
			-- -e, --echo-passphrase Echo passphrase or mnemonic to screen upon entry
			-- -i, --in-fmt=      f  Input is from wallet format 'f' (see FMT CODES below)
			-- -H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
			+                        'f' at offset 'o' (comma-separated)
			-- -O, --old-incog-fmt   Specify old-format incognito input
			-- -k, --use-internal-keccak-module Force use of the internal keccak module
			-- -K, --keygen-backend=n Use backend 'n' for public key generation.  Options
			+                        for {coin_id}: {kgs}
			-- -l, --seed-len=    l  Specify wallet seed length of 'l' bits.  This option
			+                        is required only for brainwallet and incognito inputs
			+                        with non-standard (< {dsl}-bit) seed lengths.
			-- -p, --hash-preset= p  Use the scrypt hash parameters defined by preset 'p'
			+                        for password hashing (default: '{gc.dfl_hash_preset}')
			-- -z, --show-hash-presets Show information on available hash presets
			-- -P, --passwd-file= f  Get wallet passphrase from file 'f'
			-- -q, --quiet           Produce quieter output; suppress some warnings
			-- -r, --usr-randchars=n Get 'n' characters of additional randomness from user
			+                        (min={cfg.min_urandchars}, max={cfg.max_urandchars}, default={cfg.usr_randchars})
			-- -S, --stdout          Print {what} to stdout
			-- -t, --type=t          Choose address type. Options: see ADDRESS TYPES below
			+                        (default: {dmat})
			-- -U, --subwallet=   U  Generate {what} for subwallet 'U' (see SUBWALLETS
			+                        below)
			-k -V, --viewkeys        Print viewkeys, omitting secret keys
			-- -v, --verbose         Produce more verbose output
			-k -x, --b16             Print secret keys in hexadecimal too
		""",
		'notes': """

                           NOTES FOR THIS COMMAND

Address indexes are given as a comma-separated list and/or hyphen-separated
range(s).

{n_addrkey}If available, the libsecp256k1 library will be used for address generation.


                      NOTES FOR ALL GENERATOR COMMANDS

{n_sw}{n_pw}{n_bw}

{n_at}

{n_fmt}
"""
	},
	'code': {
		'options': lambda proto, help_notes, cfg, s: s.format(
			dmat      = help_notes('dfl_mmtype'),
			kgs       = help_notes('keygen_backends'),
			coin_id   = help_notes('coin_id'),
			dsl       = help_notes('dfl_seed_len'),
			pnm       = gc.proj_name,
			what      = gen_what,
			cfg       = cfg,
			gc        = gc,
		),
		'notes': lambda help_mod, help_notes, s: s.format(
			n_addrkey = note_addrkey,
			n_sw      = help_mod('subwallet')+'\n\n',
			n_pw      = help_notes('passwd')+'\n\n',
			n_bw      = help_notes('brainwallet'),
			n_fmt     = help_notes('fmt_codes'),
			n_at      = help_notes('address_types'),
		)
	}
}

cfg = Config(opts_data=opts_data, need_amt=False)

proto = cfg._proto

addr_type = MMGenAddrType(
	proto = proto,
	id_str = cfg.type or proto.dfl_mmtype,
	errmsg = f'{cfg.type!r}: invalid parameter for --type option')

if len(cfg._args) < 1:
	cfg._usage()

if cfg.keygen_backend:
	from .keygen import check_backend
	check_backend(cfg, proto, cfg.keygen_backend, cfg.type)

idxs = addrlist.AddrIdxList(fmt_str=cfg._args.pop())

from .fileutil import get_seed_file
sf = get_seed_file(cfg, nargs=1)

from .ui import do_license_msg
do_license_msg(cfg)

ss = Wallet(cfg, fn=sf)

ss_seed = ss.seed if cfg.subwallet is None else ss.seed.subseed(cfg.subwallet, print_msg=True)

if cfg.no_addresses:
	gen_clsname = 'KeyList'
elif cfg.viewkeys:
	gen_clsname = 'ViewKeyAddrList'

al = getattr(addrlist, gen_clsname)(
	cfg       = cfg,
	proto     = proto,
	seed      = ss_seed,
	addr_idxs = idxs,
	mmtype    = addr_type)

af = al.file

af.format()

if al.gen_addrs and cfg.print_checksum:
	from .util import Die
	Die(0, al.checksum)

from .ui import keypress_confirm
if al.gen_keys and keypress_confirm(cfg, 'Encrypt key list?'):
	af.encrypt()
	af.write(
		binary = True,
		desc = f'encrypted {af.desc}')
else:
	af.write()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.unsigned: Bitcoin unsigned transaction class
"""

from ....tx import unsigned as TxBase
from ....obj import CoinTxID, MMGenDict
from ....util import msg, msg_r, ymsg, suf, die
from .completed import Completed

class Unsigned(Completed, TxBase.Unsigned):
	desc = 'unsigned transaction'

	# Return signed object or False. Don’t exit or raise exception:
	async def sign(self, keys, tx_num_str=''):

		from ....exception import TransactionChainMismatch
		try:
			self.check_correct_chain()
		except TransactionChainMismatch:
			return False

		if (self.has_segwit_inputs() or self.has_segwit_outputs()) and not self.proto.cap('segwit'):
			ymsg(f"TX has Segwit inputs or outputs, but {self.coin} doesn't support Segwit!")
			return False

		self.check_pubkey_scripts()

		self.cfg._util.qmsg(f'Passing {len(keys)} key{suf(keys)} to {self.rpc.daemon.exec_fn}')

		if self.has_segwit_inputs():
			from ....addrgen import KeyGenerator, AddrGenerator
			kg = KeyGenerator(self.cfg, self.proto, 'std')
			ag = AddrGenerator(self.cfg, self.proto, 'segwit')
			keydict = MMGenDict([(d.addr, d.sec) for d in keys])

		sig_data = []
		for d in self.inputs:
			e = {k:getattr(d, k) for k in ('txid', 'vout', 'scriptPubKey', 'amt')}
			e['amount'] = e['amt']
			del e['amt']
			if d.mmtype == 'S':
				e['redeemScript'] = ag.to_segwit_redeem_script(kg.gen_data(keydict[d.addr]))
			sig_data.append(e)

		msg_r(f'Signing transaction{tx_num_str}...')
		wifs = [d.sec.wif for d in keys]

		try:
			args = (
				('signrawtransaction',       self.serialized, sig_data, wifs, self.proto.sighash_type),
				('signrawtransactionwithkey', self.serialized, wifs, sig_data, self.proto.sighash_type)
			)['sign_with_key' in self.rpc.caps]
			ret = await self.rpc.call(*args)
		except Exception as e:
			ymsg(self.rpc.daemon.sigfail_errmsg(e))
			return False

		try:
			self.update_serialized(ret['hex'])
			from ....tx import SignedTX
			new = SignedTX(cfg=self.cfg, data=self.__dict__, automount=self.automount)
			tx_decoded = await self.rpc.call('decoderawtransaction', ret['hex'])
			new.compare_size_and_estimated_size(tx_decoded)
			new.coin_txid = CoinTxID(self.deserialized.txid)
			if not new.coin_txid == tx_decoded['txid']:
				die('BadMMGenTxID', 'txid mismatch (after signing)')
			msg('OK')
			return new
		except Exception as e:
			ymsg(f'\n{e.args[0]}')
			return False

class AutomountUnsigned(TxBase.AutomountUnsigned, Unsigned):
	pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.httpd.thornode.swap: Thornode swap HTTP server
"""

import time, re, json
from wsgiref.util import request_uri

from mmgen.cfg import Config
from mmgen.amt import UniAmt
from mmgen.protocol import init_proto

from ...include.common import eth_inbound_addr, thorchain_router_addr_file

from . import ThornodeServer

cfg = Config()

# https://thornode.ninerealms.com/thorchain/quote/swap?from_asset=BCH.BCH&to_asset=LTC.LTC&amount=1000000
prices = {'BTC': 97000, 'LTC': 115, 'BCH': 330, 'ETH': 2304, 'MM1': 0.998, 'RUNE': 1.4}
gas_rate_units = {'ETH': 'gwei', 'BTC': 'satsperbyte'}
recommended_gas_rate = {'ETH': '1', 'BTC': '6'}

data_template_from_rune = {
	'outbound_delay_blocks': 0,
	'outbound_delay_seconds': 0,
	'fees': {
		'asset': 'BTC.BTC',
		'affiliate': '0',
		'outbound': '1182',
		'liquidity': '110',
		'total': '1292',
		'slippage_bps': 7,
		'total_bps': 92
	},
	'warning': 'Do not cache this response. Do not send funds after the expiry.',
	'notes': 'Broadcast a MsgDeposit to the THORChain network with the appropriate memo. Do not use multi-in, multi-out transactions.',
	'max_streaming_quantity': 0,
	'streaming_swap_blocks': 0
}

data_template_to_rune = {
	'inbound_confirmation_blocks': 2,
	'inbound_confirmation_seconds': 24,
	'outbound_delay_blocks': 0,
	'outbound_delay_seconds': 0,
	'fees': {
		'asset': 'THOR.RUNE',
		'affiliate': '0',
		'outbound': '2000000',
		'liquidity': '684966',
		'total': '2684966',
		'slippage_bps': 8,
		'total_bps': 31
	},
	'router': '0xD37BbE5744D730a1d98d8DC97c42F0Ca46aD7146',
	'warning': 'Do not cache this response. Do not send funds after the expiry.',
	'notes': 'Base Asset: Send the inbound_address the asset with the memo encoded in hex in the data field. Tokens: First approve router to spend tokens from user: asset.approve(router, amount). Then call router.depositWithExpiry(inbound_address, asset, amount, memo, expiry). Asset is the token contract address. Amount should be in native asset decimals (eg 1e18 for most tokens). Do not swap to smart contract addresses.',
	'dust_threshold': '1',
	'recommended_gas_rate': '1',
	'max_streaming_quantity': 0,
	'streaming_swap_blocks': 0,
	'total_swap_seconds': 24
}

data_template_btc = {
	'inbound_confirmation_blocks': 4,
	'inbound_confirmation_seconds': 2400,
	'outbound_delay_blocks': 5,
	'outbound_delay_seconds': 30,
	'fees': {
		'asset': 'LTC.LTC',
		'affiliate': '0',
		'outbound': '878656',
		'liquidity': '8945012',
		'total': '9823668',
		'slippage_bps': 31,
		'total_bps': 34
	},
	'warning': 'Do not cache this response. Do not send funds after the expiry.',
	'notes': 'First output should be to inbound_address, second output should be change back to self, third output should be OP_RETURN, limited to 80 bytes. Do not send below the dust threshold. Do not use exotic spend scripts, locks or address formats.',
	'dust_threshold': '10000',
	'max_streaming_quantity': 0,
	'streaming_swap_blocks': 0,
	'total_swap_seconds': 2430
}

data_template_eth = {
	'inbound_confirmation_blocks': 2,
	'inbound_confirmation_seconds': 24,
	'outbound_delay_blocks': 0,
	'outbound_delay_seconds': 0,
	'fees': {
		'asset': 'BTC.BTC',
		'affiliate': '0',
		'outbound': '1097',
		'liquidity': '77',
		'total': '1174',
		'slippage_bps': 15,
		'total_bps': 237
	},
	'router': '0xD37BbE5744D730a1d98d8DC97c42F0Ca46aD7146',
	'warning': 'Do not cache this response. Do not send funds after the expiry.',
	'notes': 'Base Asset: Send the inbound_address the asset with the memo encoded in hex in the data field. Tokens: First approve router to spend tokens from user: asset.approve(router, amount). Then call router.depositWithExpiry(inbound_address, asset, amount, memo, expiry). Asset is the token contract address. Amount should be in native asset decimals (eg 1e18 for most tokens). Do not swap to smart contract addresses.',
	'recommended_gas_rate': '1',
	'max_streaming_quantity': 0,
	'streaming_swap_blocks': 0,
	'total_swap_seconds': 24
}

def make_inbound_addr(proto, mmtype):
	if proto.is_evm:
		return '0x' + eth_inbound_addr # non-checksummed as per ninerealms thornode
	else:
		from mmgen.tool.coin import tool_cmd
		n = int(time.time()) // (60 * 60 * 24) # increments once every 24 hrs
		return tool_cmd(
			cfg     = cfg,
			cmdname = 'pubhash2addr',
			proto   = proto,
			mmtype  = mmtype).pubhash2addr(f'{n:040x}')

class ThornodeSwapServer(ThornodeServer):
	port = 18900
	name = 'thornode swap server'
	request_pat = r'/thorchain/quote/swap\?from_asset=(\S+)\.(\S+)&to_asset=(\S+)\.(\S+)&amount=(\d+)'

	def make_response_body(self, method, environ):

		m = re.search(self.request_pat, request_uri(environ))
		send_chain, send_asset, recv_chain, recv_asset, amt_atomic = m.groups()

		in_amt = UniAmt(int(amt_atomic), from_unit='satoshi')
		out_amt = in_amt * (prices[send_asset] / prices[recv_asset])

		data_template = (
			data_template_from_rune if send_asset == 'RUNE' else
			data_template_to_rune if recv_asset == 'RUNE' else
			data_template_eth if send_asset == 'ETH' else
			data_template_btc)

		data = data_template | {
			'recommended_min_amount_in': str(int(70 * 10**8 / prices[send_asset])), # $70
			'expected_amount_out': str(out_amt.to_unit('satoshi')),
			'expiry': int(time.time()) + (10 * 60),
		}

		if send_asset != 'RUNE':
			send_proto = init_proto(cfg, send_chain, network='regtest', need_amt=True)
			data.update({
				'inbound_address': make_inbound_addr(send_proto, send_proto.preferred_mmtypes[0]),
				'gas_rate_units': gas_rate_units[send_proto.base_proto_coin],
				'recommended_gas_rate': recommended_gas_rate[send_proto.base_proto_coin]
			})

		if send_asset == 'MM1':
			eth_proto = init_proto(cfg, 'eth', network='regtest')
			with open(thorchain_router_addr_file) as fh:
				raw_addr = fh.read().strip()
			data['router'] = '0x' + eth_proto.checksummed_addr(raw_addr)

		return json.dumps(data).encode()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help.txcreate: txcreate and txdo help notes for the MMGen Wallet suite
"""

def help(proto, cfg):
	outputs_info = (
	"""
Outputs are specified in the form ADDRESS,AMOUNT or ADDRESS.  The first form
creates an output sending the given amount to the given address.  The bare
address form designates the given address as either the change output or the
sole output of the transaction (excluding any data output).  Exactly one bare
address argument is required.

For convenience, the bare address argument may be given as ADDRTYPE_CODE or
SEED_ID:ADDRTYPE_CODE (see ADDRESS TYPES below). In the first form, the first
unused address of type ADDRTYPE_CODE for each Seed ID in the tracking wallet
will be displayed in a menu, with the user prompted to select one.  In the
second form, the user specifies the Seed ID as well, allowing the script to
select the transaction’s change output or single output without prompting.
See EXAMPLES below.

A single DATA_SPEC argument may also be given on the command line to create
an OP_RETURN data output with a zero spend amount.  This is the preferred way
to embed data in the blockchain.  DATA_SPEC may be of the form "data":DATA
or "hexdata":DATA. In the first form, DATA is a string in your system’s native
encoding, typically UTF-8.  In the second, DATA is a hexadecimal string (with
the leading ‘0x’ omitted) encoding the binary data to be embedded.  In both
cases, the resulting byte string must not exceed {bl} bytes in length.
""".format(bl=proto.max_op_return_data_len)
	if proto.base_proto == 'Bitcoin' else """
The transaction output is specified in the form ADDRESS,AMOUNT.
""")

	fee_info = """
If the transaction fee is not specified on the command line (see FEE
SPECIFICATION below), it will be calculated dynamically using network fee
estimation for the default (or user-specified) number of confirmations.
If network fee estimation fails, the user will be prompted for a fee.

Network-estimated fees will be multiplied by the value of --fee-adjust, if
specified.
""" if proto.has_usr_fee else ''

	return f"""
The transaction’s outputs are listed on the command line, while its inputs
are chosen from a list of the wallet’s unspent outputs via an interactive
menu.  Alternatively, inputs may be specified using the --inputs option.

Addresses on the command line can be either native coin addresses or MMGen
IDs in the form SEED_ID:ADDRTYPE_CODE:INDEX.
{outputs_info}{fee_info}"""



import os as overlay_fake_os
from .util_orig import *

if overlay_fake_os.getenv('MMGEN_TEST_SUITE_DETERMINISTIC'):

	class overlay_fake_data:
		make_timestamp = make_timestamp
		make_timestr = make_timestr
		time_iter = (1862651471 + (i*60) for i in range(1000000)) # 9 Jan 2029 11:11:11

	def make_timestamp(secs=None):
		return overlay_fake_data.make_timestamp(next(overlay_fake_data.time_iter))

	def make_timestr(secs=None):
		return overlay_fake_data.make_timestr(next(overlay_fake_data.time_iter))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.keys: transaction keys class for the MMGen suite
"""

from ..cfg import gc
from ..util import msg, suf, fmt, die, remove_dups, get_extension
from ..addr import MMGenAddrType
from ..addrlist import AddrIdxList, KeyAddrList
from ..wallet import Wallet, get_wallet_extensions, get_wallet_cls

def _pop_matching_fns(args, cmplist): # strips found args
	return list(reversed(
		[args.pop(args.index(a)) for a in reversed(args) if get_extension(a) in cmplist]))

def pop_txfiles(cfg):
	from .unsigned import Unsigned, AutomountUnsigned
	ext = (AutomountUnsigned if cfg.autosign else Unsigned).ext
	ret = _pop_matching_fns(cfg._args, [ext])
	if not ret:
		die(1, f'You must specify a raw transaction file with extension ‘.{ext}’')
	return ret

def pop_seedfiles(cfg, *, ignore_dfl_wallet=False, empty_ok=False):
	# favor unencrypted seed sources first, as they don't require passwords
	ret = _pop_matching_fns(cfg._args, get_wallet_extensions('unenc'))
	from ..filename import find_file_in_dir
	if not ignore_dfl_wallet: # Make this the first encrypted ss in the list
		if wf := find_file_in_dir(get_wallet_cls('mmgen'), cfg.data_dir):
			ret.append(wf)
	ret += _pop_matching_fns(cfg._args, get_wallet_extensions('enc'))
	if not (ret
			or empty_ok
			or cfg.mmgen_keys_from_file
			or cfg.keys_from_file):
		die(1, 'You must specify a seed or key source!')
	return ret

def get_keylist(cfg):
	if cfg.keys_from_file:
		from ..fileutil import get_lines_from_file
		return get_lines_from_file(
			cfg,
			cfg.keys_from_file,
			desc = 'key-address data',
			trim_comments = True)

def get_keyaddrlist(cfg, proto):
	if cfg.mmgen_keys_from_file:
		return KeyAddrList(cfg, proto, infile=cfg.mmgen_keys_from_file)

class TxKeys:
	"""
	produce a list of keys to be used for transaction signing

	Keys are taken from a key-address list and/or flat key list, and/or generated
	from provided seed sources.  Seed sources are searched for subseeds to allow
	signing of inputs with subseed addresses by parent seeds.

	In addition, MMGenID-to-address mappings are verified for output addresses, as
	well as the swap destination address, if applicable.  For this reason, seeds or
	key-address files for all MMGen addresses involved in the transaction must be
	provided.

	Verification of the swap memo against TX metadata is also performed.
	"""
	def __init__(
			self,
			cfg,
			tx,
			*,
			seedfiles   = None,
			keylist     = None,
			keyaddrlist = None,
			passwdfile  = None,
			autosign    = False):
		self.cfg         = cfg
		self.tx          = tx
		self.seedfiles   = seedfiles or pop_seedfiles(cfg)
		self.keylist     = keylist if autosign else keylist or get_keylist(cfg)
		self.keyaddrlist = keyaddrlist if autosign else keyaddrlist or get_keyaddrlist(cfg, tx.proto)
		self.passwdfile  = passwdfile
		self.autosign    = autosign
		self.saved_seeds = {}

	def get_keys_for_non_mmgen_inputs(self):
		err_fs = 'ERROR: a key file must be supplied for the following non-{} address{}:{}'
		sep = '\n    '
		if addrs := self.tx.get_non_mmaddrs('inputs'):
			self.tx.check_non_mmgen_inputs(
				caller = 'autosign' if self.autosign and self.keylist else 'txsign',
				non_mmaddrs = addrs)
			kal = KeyAddrList(
				cfg         = self.cfg,
				proto       = self.tx.proto,
				addrlist    = addrs,
				skip_chksum = True)
			if self.keylist:
				kal.add_wifs(self.keylist)
			if missing := kal.list_missing('sec'):
				die(2, err_fs.format(gc.proj_name, suf(missing, 'es'), sep + sep.join(missing)))
			return kal.data
		else:
			return []

	def get_seed_for_seed_id(self, sid):

		if sid in self.saved_seeds:
			return self.saved_seeds[sid]

		subseeds_checked = False
		seed = None

		while True:
			if self.seedfiles:
				seed = Wallet(
					self.cfg,
					fn = self.seedfiles.pop(0),
					ignore_in_fmt = True,
					passwd_file = self.passwdfile).seed
			elif self.saved_seeds and subseeds_checked is False:
				seed = self.saved_seeds[list(self.saved_seeds)[0]].subseed_by_seed_id(sid, print_msg=True)
				subseeds_checked = True
				if not seed:
					continue
			elif self.cfg.in_fmt:
				self.cfg._util.qmsg(f'Need seed data for Seed ID {sid}')
				seed = Wallet(self.cfg, passwd_file=self.passwdfile).seed
				msg(f'User input produced Seed ID {seed.sid}')
				if not seed.sid == sid: # TODO: add test
					seed = seed.subseed_by_seed_id(sid, print_msg=True)

			if seed:
				self.saved_seeds[seed.sid] = seed
				if seed.sid == sid:
					return seed
			else:
				die(2, f'ERROR: No seed source found for Seed ID: {sid}')

	def generate_kals_for_mmgen_addrs(self, need_keys, proto):
		mmids = [e.mmid for e in need_keys]
		sids = remove_dups((i.sid for i in mmids), quiet=True)
		self.cfg._util.vmsg('Need seed{}: {}'.format(suf(sids), ' '.join(sids)))
		for sid in sids:
			seed = self.get_seed_for_seed_id(sid) # raises exception if seed not found
			for id_str in MMGenAddrType.mmtypes:
				idx_list = [i.idx for i in mmids if i.sid == sid and i.mmtype == id_str]
				if idx_list:
					yield KeyAddrList(
						cfg         = self.cfg,
						proto       = proto,
						seed        = seed,
						addr_idxs   = AddrIdxList(idx_list=idx_list),
						mmtype      = MMGenAddrType(proto, id_str),
						skip_chksum = True)

	def add_keys(self, src, io_list, *, from_keyaddrlist=False):

		if not (need_keys := [e for e in io_list if e.mmid and not e.have_wif]):
			return []

		proto = need_keys[0].proto

		if from_keyaddrlist:
			desc = 'key-address file'
			err_desc = 'From key-address file:'
			kals = tuple([self.keyaddrlist])
		else:
			desc = 'seed(s)'
			err_desc = 'Generated from seed:'
			kals = tuple(self.generate_kals_for_mmgen_addrs(need_keys, proto))

		self.cfg._util.qmsg(
			f'Checking {gc.proj_name} -> {proto.coin} address mappings for {src} (from {desc})')

		def gen_keys():
			for e in need_keys:
				for kal in kals:
					for f in kal.data:
						if mmid := f'{kal.al_id}:{f.idx}' == e.mmid:
							if f.addr == e.addr:
								e.have_wif = True
								if src == 'inputs':
									yield f
							else:
								die(3, fmt(f"""
									{gc.proj_name} -> {proto.coin} address mappings differ!
									{err_desc:<23} {mmid} -> {f.addr}
									{'tx file:':<23} {e.mmid} -> {e.addr}
									""").strip())

		if new_keys := list(gen_keys()):
			self.cfg._util.vmsg(f'Added {len(new_keys)} wif key{suf(new_keys)} from {desc}')

		return new_keys

	@property
	def keys(self):
		"""
		produce a list of signing keys and perform checks on output and swap destination
		addresses
		"""
		ret = self.get_keys_for_non_mmgen_inputs()
		memo_output = self.tx.check_swap_memo() # do this for non-swap transactions too!

		if self.keyaddrlist:
			ret += self.add_keys('inputs', self.tx.inputs, from_keyaddrlist=True)
			self.add_keys('outputs', self.tx.outputs, from_keyaddrlist=True)
			if memo_output:
				self.add_keys('swap destination address', [memo_output], from_keyaddrlist=True)

		ret += self.add_keys('inputs', self.tx.inputs)
		self.add_keys('outputs', self.tx.outputs)
		if memo_output:
			self.add_keys('swap destination address', [memo_output])

		# this (boolean) attr isn't needed in transaction file
		self.tx.delete_attrs('inputs', 'have_wif')
		self.tx.delete_attrs('outputs', 'have_wif')

		if extra_sids := remove_dups(
				(s for s in self.saved_seeds
					if s not in self.tx.get_sids('inputs') + self.tx.get_sids('outputs')),
				quiet = True):
			msg('Unused Seed ID{}: {}'.format(suf(extra_sids), ' '.join(extra_sids)))

		return ret



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.xmr.params: Monero protocol
"""

from collections import namedtuple

from ...protocol import CoinProtocol, _nw
from ...obj import HexStr

parsed_addr = namedtuple('parsed_addr', ['ver_bytes', 'data', 'payment_id'])

class MoneroViewKey(HexStr):
	color, width, hexcase = 'cyan', 64, 'lower' # FIXME - no checking performed

# https://github.com/monero-project/monero/blob/master/src/cryptonote_config.h
class mainnet(CoinProtocol.RPC, CoinProtocol.DummyWIF, CoinProtocol.Base):

	network_names  = _nw('mainnet', 'stagenet', None)
	base_proto     = 'Monero'
	base_proto_coin = 'XMR'
	base_coin      = 'XMR'
	addr_ver_info  = {'12': 'monero', '2a': 'monero_sub', '13': 'monero_integrated'}
	pubkey_types   = ('monero',)
	mmtypes        = ('M',)
	dfl_mmtype     = 'M'
	pubkey_type    = 'monero' # required by DummyWIF
	avg_bdi        = 120
	privkey_len    = 32
	mmcaps         = ('rpc',)
	coin_amt       = 'XMRAmt'
	sign_mode      = 'standalone'

	coin_cfg_opts = (
		'ignore_daemon_version',
		'rpc_port',
	)

	def get_addr_len(self, addr_fmt):
		return (64, 72)[addr_fmt == 'monero_integrated']

	def preprocess_key(self, sec, pubkey_type): # reduce key
		from ...contrib.ed25519 import l
		return int.to_bytes(
			int.from_bytes(sec[::-1], 'big') % l,
			self.privkey_len,
			'big')[::-1]

	def decode_addr(self, addr):

		from ...baseconv import baseconv

		def b58dec(addr_str):
			bc = baseconv('b58')
			l = len(addr_str)
			a = b''.join([bc.tobytes(addr_str[i*11:i*11+11], pad=8) for i in range(l//11)])
			b = bc.tobytes(addr_str[-(l%11):], pad=5)
			return a + b

		ret = b58dec(addr)

		chk = self.keccak_256(ret[:-4]).digest()[:4]

		assert ret[-4:] == chk, f'{ret[-4:].hex()}: incorrect checksum.  Correct value: {chk.hex()}'

		return self.decode_addr_bytes(ret[:-4])

	def parse_addr(self, ver_bytes, addr_bytes, fmt):
		addr_len = self.get_addr_len('monero')
		return parsed_addr(
			ver_bytes  = ver_bytes,
			data       = addr_bytes[:addr_len],
			payment_id = addr_bytes[addr_len:] if fmt == 'monero_integrated' else None,
		)

	def pubhash2addr(self, *args, **kwargs):
		raise NotImplementedError('Monero addresses do not support pubhash2addr()')

	def viewkey(self, viewkey_str):
		return MoneroViewKey.__new__(MoneroViewKey, viewkey_str)

class testnet(mainnet): # use stagenet for testnet
	# testnet is {'35', '3f', '36'}
	addr_ver_info = {'18': 'monero', '24': 'monero_sub', '19': 'monero_integrated'}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.include.runner: test runner for the MMGen Wallet cmdtest suite
"""

import sys, os, time, asyncio

from mmgen.cfg import gc
from mmgen.color import red, yellow, green, blue, cyan, gray, nocolor
from mmgen.util import msg, Msg, rmsg, ymsg, bmsg, die, suf, make_timestr

from ...include.common import (
	cmdtest_py_log_fn,
	iqmsg,
	omsg,
	omsg_r,
	ok,
	start_test_daemons,
	init_coverage,
	clean
)

from .common import get_file_with_ext, confirm_continue
from .cfg import cfgs, cmd_groups_dfl
from .group_mgr import CmdGroupMgr

def format_args(args):
	try:
		return ' '.join((f"'{a}'" if ' ' in a else a) for a in args).replace('\\', '/') # for MSYS2
	except Exception as e:
		print(type(e), e)
		print('cmdline:', args)

class CmdTestRunner:
	'cmdtest.py test runner'

	def __del__(self):
		if self.logging:
			self.log_fd.close()

	def __init__(self, cfg, repo_root, data_dir, trash_dir, trash_dir2):

		self.cfg = cfg
		self.proto = cfg._proto
		self.data_dir = data_dir
		self.trash_dir = trash_dir
		self.trash_dir2 = trash_dir2
		self.cmd_total = 0
		self.rebuild_list = {}
		self.gm = CmdGroupMgr(cfg)
		self.repo_root = repo_root
		self.warnings = []
		self.skipped_warnings = []
		self.resume_cmd = None
		self.deps_only = None
		self.logging = self.cfg.log or os.getenv('MMGEN_EXEC_WRAPPER')
		self.testing_segwit = cfg.segwit or cfg.segwit_random or cfg.bech32
		self.network_id = self.proto.coin.lower() + ('_tn' if self.proto.testnet else '')
		self.daemon_started = False
		self.quiet = not (cfg.exact_output or cfg.verbose)

		global qmsg, qmsg_r
		if cfg.exact_output:
			qmsg = qmsg_r = lambda s: None
		else:
			qmsg = cfg._util.qmsg
			qmsg_r = cfg._util.qmsg_r

		if self.logging:
			self.log_fd = open(cmdtest_py_log_fn, 'a')
			self.log_fd.write(f'\nLog started: {make_timestr()} UTC\n')
			omsg(f'INFO → Logging to file {cmdtest_py_log_fn!r}')
		else:
			self.log_fd = None

		if self.cfg.coverage:
			coverdir, accfile = init_coverage()
			omsg(f'INFO → Writing coverage files to {coverdir!r}')
			self.pre_args = ['python3', '-m', 'trace', '--count', '--coverdir='+coverdir, '--file='+accfile]
		else:
			self.pre_args = ['python3'] if sys.platform == 'win32' else []

		if self.cfg.pexpect_spawn:
			omsg('INFO → Using pexpect.spawn() for real terminal emulation')

		self.set_spawn_env()
		self.start_time = time.time()

	def do_between(self):
		if self.cfg.pause:
			confirm_continue()
		elif not (self.quiet or self.cfg.skipping_deps):
			sys.stderr.write('\n')

	def set_spawn_env(self):

		self.spawn_env = dict(os.environ)
		self.spawn_env.update({
			'MMGEN_NO_LICENSE': '1',
			'MMGEN_BOGUS_SEND': '1',
			'MMGEN_TEST_SUITE_PEXPECT': '1',
			'EXEC_WRAPPER_DO_RUNTIME_MSG':'1',
			# if cmdtest.py itself is running under exec_wrapper, disable writing of traceback file for spawned script
			'EXEC_WRAPPER_TRACEBACK': '' if os.getenv('MMGEN_EXEC_WRAPPER') else '1',
		})

		if self.cfg.exact_output:
			from mmgen.term import get_terminal_size
			self.spawn_env['MMGEN_COLUMNS'] = str(get_terminal_size().width)
		else:
			self.spawn_env['MMGEN_COLUMNS'] = '120'

	def spawn_wrapper(
			self,
			cmd             = '',
			args            = [],
			extra_desc      = '',
			no_output       = False,
			msg_only        = False,
			log_only        = False,
			no_msg          = False,
			cmd_dir         = 'cmds',
			no_exec_wrapper = False,
			timeout         = None,
			pexpect_spawn   = None,
			direct_exec     = False,
			no_passthru_opts = False,
			spawn_env_override = None,
			exit_val        = None,
			silent          = False,
			env             = {}):

		self.exit_val = exit_val

		desc = self.tg.test_name if self.cfg.names else self.gm.dpy_data[self.tg.test_name][1]
		if extra_desc:
			desc += ' ' + extra_desc

		cmd_path = (
			cmd if self.cfg.system # self.cfg.system is broken for main test group with overlay tree
			else os.path.relpath(os.path.join(self.repo_root, cmd_dir, cmd)))

		passthru_opts = (
			self.passthru_opts if not no_passthru_opts else
			[] if no_passthru_opts is True else
			[o for o in self.passthru_opts
				if o[2:].split('=')[0].replace('-','_') not in no_passthru_opts])

		args = (
			self.pre_args +
			([] if no_exec_wrapper else ['scripts/exec_wrapper.py']) +
			[cmd_path] +
			passthru_opts +
			args)

		cmd_disp = format_args(args)

		if self.logging:
			self.log_fd.write('[{}][{}:{}] {}\n'.format(
				(self.proto.coin.lower() if 'coin' in self.tg.passthru_opts else 'NONE'),
				self.tg.group_name,
				self.tg.test_name,
				cmd_disp))

		if log_only:
			return

		for i in args: # die only after writing log entry
			if not isinstance(i, str):
				die(2, 'Error: missing input files in cmd line?:\nName: {}\nCmdline: {!r}'.format(
					self.tg.test_name,
					args))

		if not no_msg:
			t_pfx = '' if self.cfg.no_timings else f'[{time.time() - self.start_time:08.2f}] '
			if (not self.quiet) or self.cfg.print_cmdline:
				omsg(green(f'{t_pfx}Testing: {desc}'))
				if not msg_only:
					clr1, clr2 = (nocolor, nocolor) if self.cfg.print_cmdline else (green, cyan)
					omsg(
						clr1('Executing: ') +
						clr2(repr(cmd_disp) if sys.platform == 'win32' else cmd_disp)
					)
			else:
				omsg_r('{a}Testing {b}: {c}'.format(
					a = t_pfx,
					b = desc,
					c = 'OK\n' if direct_exec or self.cfg.direct_exec else ''))

		if msg_only:
			return

		# NB: the `pexpect_spawn` arg enables hold_protect and send_delay while the corresponding cmdline
		# option does not.  For performance reasons, this is the desired behavior.  For full emulation of
		# the user experience with hold protect enabled, specify --buf-keypress or --demo.
		send_delay = 0.4 if pexpect_spawn is True or self.cfg.buf_keypress else None
		pexpect_spawn = pexpect_spawn if pexpect_spawn is not None else bool(self.cfg.pexpect_spawn)

		spawn_env = dict(spawn_env_override or self.tg.spawn_env)
		spawn_env.update({
			'MMGEN_HOLD_PROTECT_DISABLE': '' if send_delay else '1',
			'MMGEN_TEST_SUITE_POPEN_SPAWN': '' if pexpect_spawn else '1',
			'EXEC_WRAPPER_EXIT_VAL': '' if exit_val is None else str(exit_val),
		})
		spawn_env.update(env)

		from .pexpect import CmdTestPexpect
		return CmdTestPexpect(
			args          = args,
			no_output     = no_output,
			spawn_env     = spawn_env,
			pexpect_spawn = pexpect_spawn,
			timeout       = timeout,
			send_delay    = send_delay,
			silent        = silent,
			direct_exec   = direct_exec)

	def end_msg(self):
		t = int(time.time() - self.start_time)
		sys.stderr.write(green(
			f'{self.cmd_total} test{suf(self.cmd_total)} performed' +
			('\n' if self.cfg.no_timings else f'.  Elapsed time: {t//60:02d}:{t%60:02d}\n')
		))

	def init_group(self, gname, sg_name=None, cmd=None, quiet=False, do_clean=True):

		from .cfg import cmd_groups_altcoin
		if self.cfg.no_altcoin and gname in cmd_groups_altcoin:
			omsg(gray(f'INFO → skipping test group {gname!r} (--no-altcoin)'))
			return None

		ct_cls = self.gm.load_mod(gname)

		if sys.platform in ct_cls.platform_skip:
			omsg(gray(f'INFO → skipping test {gname!r} for platform {sys.platform!r}'))
			return None

		for k in ('segwit', 'segwit_random', 'bech32'):
			if getattr(self.cfg, k):
				segwit_opt = k
				break
		else:
			segwit_opt = None

		def gen_msg():
			yield ('{g}:{c}' if cmd else 'test group {g!r}').format(g=gname, c=cmd)
			if len(ct_cls.networks) != 1:
				yield f' for {self.proto.coin} {self.proto.network}'
			if segwit_opt:
				yield ' (--{})'.format(segwit_opt.replace('_', '-'))

		m = ''.join(gen_msg())

		if segwit_opt and not ct_cls.segwit_opts_ok:
			iqmsg(gray(f'INFO → skipping {m}'))
			return None

		# 'networks = ()' means all networks allowed
		nws = [(e.split('_')[0], 'testnet') if '_' in e else (e, 'mainnet') for e in ct_cls.networks]
		if nws:
			coin = self.proto.coin.lower()
			for a, b in nws:
				if a == coin and b == self.proto.network:
					break
			else:
				iqmsg(gray(f'INFO → skipping {m} for {self.proto.coin} {self.proto.network}'))
				return None

		if do_clean and not self.cfg.skipping_deps:
			clean(
				cfgs,
				tmpdir_ids = ct_cls.tmpdir_nums,
				extra_dirs = [self.data_dir, self.trash_dir, self.trash_dir2])

		if not quiet:
			bmsg('Executing ' + m)

		if (not self.daemon_started) and self.gm.get_cls_by_gname(gname).need_daemon:
			start_test_daemons(self.network_id, remove_datadir=True)
			self.daemon_started = True

		if hasattr(self, 'tg'):
			del self.tg

		self.tg = self.gm.gm_init_group(self.cfg, self, gname, sg_name, self.spawn_wrapper)
		self.ct_clsname = type(self.tg).__name__

		# pass through opts from cmdline (po.user_opts)
		self.passthru_opts = ['--{}{}'.format(
				k.replace('_', '-'),
				'' if self.cfg._uopts[k] is True else '=' + self.cfg._uopts[k]
			) for k in self.cfg._uopts
				if self.cfg._uopts[k] and k in self.tg.base_passthru_opts + self.tg.passthru_opts]

		if self.cfg.resuming:
			rc = self.cfg.resume or self.cfg.resume_after
			offset = 1 if self.cfg.resume_after else 0
			self.resume_cmd = self.gm.cmd_list[self.gm.cmd_list.index(rc)+offset]
			omsg(f'INFO → Resuming at command {self.resume_cmd!r}')
			if self.cfg.step:
				self.cfg.exit_after = self.resume_cmd

		if self.cfg.exit_after and self.cfg.exit_after not in self.gm.cmd_list:
			die(1, f'{self.cfg.exit_after!r}: command not recognized')

		return self.tg

	def run_tests(self, cmd_args):
		gname_save = None

		def parse_arg(arg):
			if '.' in arg:
				a, b = arg.split('.')
				return [a] + b.split(':') if ':' in b else [a, b, None]
			elif ':' in arg:
				a, b = arg.split(':')
				return [a, None, b]
			else:
				return [self.gm.find_cmd_in_groups(arg), None, arg]

		if cmd_args:
			for arg in cmd_args:
				if arg in self.gm.cmd_groups:
					if self.init_group(arg):
						for cmd in self.gm.cmd_list:
							self.check_needs_rerun(cmd, build=True)
							self.do_between()
				else:
					gname, sg_name, cmdname = parse_arg(arg)
					if gname:
						same_grp = gname == gname_save # same group as previous cmd: don't clean, suppress blue msg
						if self.init_group(gname, sg_name, cmdname, quiet=same_grp, do_clean=not same_grp):
							if cmdname:
								if self.cfg.deps_only:
									self.deps_only = cmdname
								try:
									self.check_needs_rerun(cmdname, build=True)
								except Exception as e: # allow calling of functions not in cmd_group
									if isinstance(e, KeyError) and e.args[0] == cmdname:
										ret = getattr(self.tg, cmdname)()
										if type(ret).__name__ == 'coroutine':
											ret = asyncio.run(ret)
										self.process_retval(cmdname, ret)
									else:
										raise
								self.do_between()
							else:
								for cmd in self.gm.cmd_list:
									self.check_needs_rerun(cmd, build=True)
									self.do_between()
							gname_save = gname
					else:
						die(1, f'{arg!r}: command not recognized')
		else:
			if self.cfg.exclude_groups:
				exclude = self.cfg.exclude_groups.split(',')
				for e in exclude:
					if e not in cmd_groups_dfl:
						die(1, f'{e!r}: group not recognized')
			for gname in cmd_groups_dfl:
				if self.cfg.exclude_groups and gname in exclude:
					continue
				if self.init_group(gname):
					for cmd in self.gm.cmd_list:
						self.check_needs_rerun(cmd, build=True)
						self.do_between()

		self.end_msg()

	def check_needs_rerun(
			self,
			cmd,
			build        = False,
			root         = True,
			force_delete = False,
			dpy          = False):

		self.tg.test_name = cmd

		if self.ct_clsname == 'CmdTestMain' and self.testing_segwit and cmd not in self.tg.segwit_do:
			return False

		rerun = root # force_delete is not passed to recursive call

		fns = []
		if force_delete or not root:
			# does cmd produce a needed dependency(ies)?
			ret = self.get_num_exts_for_cmd(cmd)
			if ret:
				for ext in ret[1]:
					fn = get_file_with_ext(cfgs[ret[0]]['tmpdir'], ext, delete=build)
					if fn:
						if force_delete:
							os.unlink(fn)
						else: fns.append(fn)
					else: rerun = True

		fdeps = self.generate_file_deps(cmd)
		cdeps = self.generate_cmd_deps(fdeps)

		for fn in fns:
			my_age = os.stat(fn).st_mtime
			for num, ext in fdeps:
				f = get_file_with_ext(cfgs[num]['tmpdir'], ext, delete=build)
				if f and os.stat(f).st_mtime > my_age:
					rerun = True

		for cdep in cdeps:
			if self.check_needs_rerun(cdep, build=build, root=False, dpy=cmd):
				rerun = True

		if build:
			if rerun:
				for fn in fns:
					if not root:
						os.unlink(fn)
				if not (dpy and self.cfg.skipping_deps):
					self.run_test(cmd)
				if not root:
					self.do_between()
		else:
			# If prog produces multiple files:
			if cmd not in self.rebuild_list or rerun is True:
				self.rebuild_list[cmd] = (rerun, fns[0] if fns else '') # FIX

		return rerun

	def run_test(self, cmd, sub=False):

		if self.deps_only and cmd == self.deps_only:
			sys.exit(0)

		if self.tg.full_data:
			d = [(str(num), ext) for exts, num in self.gm.dpy_data[cmd][2] for ext in exts]
			# delete files depended on by this cmd
			arg_list = [get_file_with_ext(cfgs[num]['tmpdir'], ext) for num, ext in d]

			# remove shared_deps from arg list
			if hasattr(self.tg, 'shared_deps'):
				arg_list = arg_list[:-len(self.tg.shared_deps)]
		else:
			arg_list = []

		if self.resume_cmd:
			if cmd != self.resume_cmd:
				return
			bmsg(f'Resuming at {self.resume_cmd!r}')
			self.resume_cmd = None
			self.cfg.skipping_deps = False
			self.cfg.resuming = False

		if self.cfg.profile:
			start = time.time()

		self.tg.test_name = cmd # NB: Do not remove, this needs to be set twice

		if self.tg.full_data:
			tmpdir_num = self.gm.dpy_data[cmd][0]
			self.tg.tmpdir_num = tmpdir_num
			for k in (test_cfg := cfgs[str(tmpdir_num)]):
				if k in self.gm.cfg_attrs:
					setattr(self.tg, k, test_cfg[k])

		ret = getattr(self.tg, cmd)(*arg_list) # run the test
		if sub:
			return ret

		if type(ret).__name__ == 'coroutine':
			ret = asyncio.run(ret)

		self.process_retval(cmd, ret)

		if self.cfg.profile:
			omsg('\r\033[50C{:.4f}'.format(time.time() - start))

		if cmd == self.cfg.exit_after:
			sys.exit(0)

	def print_warnings(self):
		if self.skipped_warnings:
			print(yellow('The following tests were skipped and may require attention:'))
			r = '-' * 72 + '\n'
			print(r+('\n'+r).join(self.skipped_warnings))
		if self.warnings:
			print(yellow('The following issues were encountered and may require attention:'))
			r = '-' * 72 + '\n'
			print(r+('\n'+r).join(self.warnings))

	def process_retval(self, cmd, ret):
		if type(ret).__name__ == 'CmdTestPexpect':
			ret.ok(exit_val=self.exit_val)
			self.cmd_total += 1
		elif ret == 'ok':
			ok()
			self.cmd_total += 1
		elif ret in ('skip', 'skip_msg', 'silent'):
			if ret == 'silent':
				self.cmd_total += 1
			elif ret == 'skip_msg':
				ok('SKIP')
		elif ret == 'error':
			die(2, red(f'\nTest {self.tg.test_name!r} failed'))
		elif isinstance(ret, tuple) and ret[0] == 'skip_warn':
			wmsg = 'Test {!r} was skipped:\n  {}'.format(cmd, '\n  '.join(ret[1].split('\n')))
			self.skipped_warnings.append(wmsg)
			if self.logging:
				self.log_fd.write(f'WARNING: {wmsg}\n')
		else:
			die(2, f'{cmd!r} returned {ret}')

	def warn(self, text):
		ymsg(text)
		wmsg = 'Test ‘{}:{}’: {}'.format(self.tg.group_name, self.tg.test_name, text)
		self.warnings.append(wmsg)
		if self.logging:
			self.log_fd.write(f'WARNING: {wmsg}\n')

	def check_deps(self, cmds): # TODO: broken, unused
		if len(cmds) != 1:
			die(1, f'Usage: {gc.prog_name} check_deps <command>')

		cmd = cmds[0]

		if cmd not in self.gm.cmd_list:
			die(1, f'{cmd!r}: unrecognized command')

		if not self.cfg.quiet:
			omsg(f'Checking dependencies for {cmd!r}')

		self.check_needs_rerun(cmd)

		w = max(map(len, self.rebuild_list)) + 1
		for cmd in self.rebuild_list:
			c = self.rebuild_list[cmd]
			m = 'Rebuild' if (c[0] and c[1]) else 'Build' if c[0] else 'OK'
			omsg('cmd {:<{w}} {}'.format(cmd+':', m, w=w))

	def generate_file_deps(self, cmd):
		return [(str(n), e) for exts, n in self.gm.dpy_data[cmd][2] for e in exts]

	def generate_cmd_deps(self, fdeps):
		return [cfgs[str(n)]['dep_generators'][ext] for n, ext in fdeps]

	def get_num_exts_for_cmd(self, cmd):
		try:
			num = str(self.gm.dpy_data[cmd][0])
		except KeyError:
			qmsg_r(f'Missing dependency {cmd!r}')
			gname = self.gm.find_cmd_in_groups(cmd)
			if gname:
				kwargs = self.gm.cmd_groups[gname][1]
				kwargs.update({'add_dpy':True})
				self.gm.create_group(gname, None, **kwargs)
				num = str(self.gm.dpy_data[cmd][0])
				qmsg(f' found in group {gname!r}')
			else:
				qmsg(' not found in any command group!')
				raise
		dgl = cfgs[num]['dep_generators']
		if cmd in dgl.values():
			exts = [k for k in dgl if dgl[k] == cmd]
			return (num, exts)
		else:
			return None



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-addrimport: Import addresses into a MMGen coin daemon tracking wallet
"""

from collections import namedtuple

from .cfg import gc, Config
from .util import msg, suf, die, fmt, async_run
from .addrlist import AddrList, KeyAddrList

opts_data = {
	'text': {
		'desc': f'Import addresses into an {gc.proj_name} tracking wallet',
		'usage':'[opts] [MMGen address file]',
		'options': """
-h, --help         Print this help message
--, --longhelp     Print help message for long (global) options
-a, --address=a    Import the single coin address 'a'
-b, --batch        Import all addresses in one RPC call
-l, --addrlist     Address source is a flat list of non-MMGen coin addresses
-k, --keyaddr-file Address source is a key-address file
-q, --quiet        Suppress warnings
-r, --rescan       Update address balances by selectively rescanning the
                   blockchain for unspent outputs that include the imported
                   address(es).  Required if any of the imported addresses
                   are already in the blockchain and have a balance.
-t, --token-addr=A Import addresses for ERC20 token with address 'A'
""",
	'notes': """

This command can also be used to update the comment fields or balances of
addresses already in the tracking wallet.

Rescanning now uses the ‘scantxoutset’ RPC call and a selective scan of
blocks containing the relevant UTXOs for much faster performance than the
previous implementation.  The rescan operation typically takes around two
minutes total, independent of the number of addresses imported.

Bear in mind that the UTXO scan will not find historical transactions: to add
them to the tracking wallet, you must perform a full or partial rescan of the
blockchain with the ‘mmgen-tool rescan_blockchain’ utility.  A full rescan of
the blockchain may take up to several hours.

It’s recommended to use ‘--rpc-backend=aio’ with ‘--rescan’.
"""
	}
}

addrimport_msgs = {
	'rescan': """
		WARNING: You’ve chosen the ‘--rescan’ option.  Rescanning the blockchain is
		necessary only if an address you’re importing is already in an output or
		outputs in the blockchain but not all transactions involving the address
		are known to the tracking wallet.

		Rescanning is performed via the UTXO method, which is only minimally affected
		by the number of addresses imported and typically takes just a few minutes.
	""",
	'bad_args': f"""
		You must specify either an {gc.proj_name} address file, a single address with
		the ‘--address’ option, or a flat list of non-{gc.proj_name} addresses with
		the ‘--addrlist’ option.
	"""
}

def parse_cmd_args(cmd_args):

	def import_mmgen_list(infile):
		return (AddrList, KeyAddrList)[bool(cfg.keyaddr_file)](cfg, proto, infile=infile)

	if len(cmd_args) == 1:
		infile = cmd_args[0]
		from .fileutil import check_infile, get_lines_from_file
		check_infile(infile)
		if cfg.addrlist:
			al = AddrList(
				cfg      = cfg,
				proto    = proto,
				addrlist = get_lines_from_file(
					cfg,
					infile,
					desc = f'non-{gc.proj_name} addresses',
					trim_comments = True))
		else:
			al = import_mmgen_list(infile)
	elif len(cmd_args) == 0 and cfg.address:
		al = AddrList(cfg, proto=proto, addrlist=[cfg.address])
		infile = 'command line'
	else:
		die(1, addrimport_msgs['bad_args'])

	return al, infile

def check_opts(twctl):
	batch = bool(cfg.batch)
	rescan = bool(cfg.rescan)

	if rescan and not 'rescan' in twctl.caps:
		msg(f"‘--rescan’ ignored: not supported by {type(twctl).__name__}")
		rescan = False

	if rescan and not cfg.quiet:
		from .ui import keypress_confirm
		keypress_confirm(
			cfg,
			f'\n{addrimport_msgs["rescan"]}\n\nContinue?',
			default_yes = True,
			do_exit = True)

	if batch and not 'batch' in twctl.caps:
		msg(f"‘--batch’ ignored: not supported by {type(twctl).__name__}")
		batch = False

	return batch, rescan

async def main():
	from .tw.ctl import TwCtl

	twctl = await TwCtl(
		cfg        = cfg,
		proto      = proto,
		token_addr = cfg.token_addr,
		mode       = 'i')

	if cfg.token or cfg.token_addr:
		msg(f'Importing for token {twctl.token.hl(0)} ({twctl.token.hlc(proto.tokensym)})')

	for k, v in addrimport_msgs.items():
		addrimport_msgs[k] = fmt(v, indent='  ', strip_char='\t').rstrip()

	al, infile = parse_cmd_args(cfg._args)

	cfg._util.qmsg(
		f'OK. {al.num_addrs} addresses'
		+ (f' from Seed ID {al.al_id.sid.hl()}' if hasattr(al.al_id, 'sid') else ''))

	msg(
		f'Importing {len(al.data)} address{suf(al.data, "es")} from {infile}'
		+ (' (batch mode)' if cfg.batch else ''))

	batch, rescan = check_opts(twctl)

	def gen_args_list(al):
		_d = namedtuple('import_data', ['addr', 'twmmid', 'comment'])
		for e in al.data:
			yield _d(
				addr    = e.addr,
				twmmid  = f'{al.al_id}:{e.idx}' if e.idx else f'{proto.base_coin.lower()}:{e.addr}',
				comment = e.comment)

	args_list = list(gen_args_list(al))

	await twctl.import_address_common(args_list, batch=batch)

	if rescan:
		await twctl.rescan_addresses({a.addr for a in args_list})

	del twctl

cfg = Config(opts_data=opts_data, need_amt=False)

proto = cfg._proto

async_run(main())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.txhistory: Bitcoin base protocol tracking wallet transaction history class
"""

from collections import namedtuple

from ....tw.txhistory import TwTxHistory
from ....tw.shared import get_tw_label, TwMMGenID
from ....addr import CoinAddr
from ....util import msg, msg_r
from ....color import nocolor, red, pink, gray
from ....obj import TwComment, CoinTxID, Int

from .rpc import BitcoinTwRPC
from .view import BitcoinTwView

class BitcoinTwTransaction:

	no_address_str = '[DATA]'

	def __init__(self, *, parent, proto, rpc,
			idx,          # unique numeric identifier of this transaction in listing
			unspent_info, # addrs in wallet with balances: {'mmid': {'addr', 'comment', 'amt'}}
			mm_map,       # all addrs in wallet: ['addr', ['twmmid', 'comment']]
			tx,           # the decoded transaction data
			wallet_vouts, # list of ints - wallet-related vouts
			prevouts,     # list of (txid,vout) pairs
			prevout_txs   # decoded transaction data for prevouts
		):

		self.parent       = parent
		self.proto        = proto
		self.rpc          = rpc
		self.idx          = idx
		self.unspent_info = unspent_info
		self.tx           = tx

		def gen_prevouts_data():
			_d = namedtuple('prevout_data', ['txid', 'data'])
			for tx in prevout_txs:
				for e in prevouts:
					if e.txid == tx['txid']:
						yield _d(e.txid, tx['vout'][e.vout])

		def gen_wallet_vouts_data():
			_d = namedtuple('wallet_vout_data', ['txid', 'data'])
			txid = self.tx['txid']
			vouts = self.tx['decoded']['vout']
			for n in wallet_vouts:
				yield _d(txid, vouts[n])

		def gen_vouts_info(data):
			_d = namedtuple('vout_info', ['txid', 'coin_addr', 'twlabel', 'data'])
			def gen():
				for d in data:
					addr = (
						d.data['scriptPubKey'].get('address') or
						d.data['scriptPubKey'].get('addresses',[self.no_address_str])[0])
					yield _d(
						txid = d.txid,
						coin_addr = addr,
						twlabel = mm_map[addr] if (addr in mm_map and mm_map[addr].twmmid) else None,
						data = d.data)
			return sorted(
				gen(),
				# if address is not MMGen, ignore address and sort by TxID + vout only
				key = lambda d: (
					(d.twlabel.twmmid.sort_key if d.twlabel and d.twlabel.twmmid.type == 'mmgen' else '')
					+ '_'
					+ d.txid
					+ '{:08d}'.format(d.data['n'])
				))

		def gen_all_addrs(src):
			for e in self.vouts_info[src]:
				if e.twlabel:
					mmid = e.twlabel.twmmid
					yield (
						(mmid if mmid.type == 'mmgen' else mmid.split(':', 1)[1]) +
						('*' if mmid in self.unspent_info else '')
					)
				else:
					yield e.coin_addr

		def total(data):
			return sum(coin_amt(d.data['value']) for d in data)

		def get_best_comment():
			"""
			find the most relevant comment for tabular (squeezed) display
			"""
			def vouts_labels(src):
				return [d.twlabel.comment for d in self.vouts_info[src] if d.twlabel and d.twlabel.comment]
			ret = vouts_labels('outputs') or vouts_labels('inputs')
			return ret[0] if ret else TwComment('')

		coin_amt = self.proto.coin_amt
		# 'outputs' refers to wallet-related outputs only
		self.vouts_info = {
			'inputs':  gen_vouts_info(gen_prevouts_data()),
			'outputs': gen_vouts_info(gen_wallet_vouts_data())
		}
		self.max_addrlen = {
			'inputs':  max(len(addr) for addr in gen_all_addrs('inputs')),
			'outputs': max(len(addr) for addr in gen_all_addrs('outputs'))
		}
		self.inputs_total = total(self.vouts_info['inputs'])
		self.outputs_total = sum(coin_amt(i['value']) for i in self.tx['decoded']['vout'])
		self.wallet_outputs_total = total(self.vouts_info['outputs'])
		self.fee = self.inputs_total - self.outputs_total
		self.nOutputs = len(self.tx['decoded']['vout'])
		self.confirmations = self.tx['confirmations']
		self.comment = get_best_comment()
		self.vsize = self.tx['decoded'].get('vsize') or self.tx['decoded']['size']
		self.txid = CoinTxID(self.tx['txid'])
		# Though 'blocktime' is flagged as an “optional” field, it’s always present for transactions
		# that are in the blockchain.  However, Bitcoin Core wallet saves a record of broadcast but
		# unconfirmed transactions, e.g. replaced transactions, and the 'blocktime' field is missing
		# for these, so use 'time' as a fallback.
		self.time = self.tx.get('blocktime') or self.tx['time']
		self.time_received = self.tx.get('timereceived')

	def blockheight_disp(self, *, color):
		return (
			# old/altcoin daemons return no 'blockheight' field, so use confirmations instead
			Int(self.rpc.blockcount + 1 - self.confirmations).hl(color=color)
			if self.confirmations > 0 else None)

	def age_disp(self, age_fmt, *, width, color):
		if age_fmt == 'confs':
			ret_str = str(self.confirmations).ljust(width)
			return gray(ret_str) if self.confirmations < 0 and color else ret_str
		elif age_fmt == 'block':
			ret = (self.rpc.blockcount - (abs(self.confirmations) - 1)) * (-1 if self.confirmations < 0 else 1)
			ret_str = str(ret).ljust(width)
			return gray(ret_str) if ret < 0 and color else ret_str
		else:
			return self.parent.date_formatter[age_fmt](self.rpc, self.tx.get('blocktime', 0))

	def txdate_disp(self, age_fmt):
		return self.parent.date_formatter[age_fmt](self.rpc, self.time)

	def txid_disp(self, *, color, width=None):
		return self.txid.hl(color=color) if width is None else self.txid.truncate(width, color=color)

	def vouts_list_disp(self, src, color, indent, addr_view_pref):

		fs1, fs2 = {
			'inputs':  ('{i},{n} {a} {A}', '{i},{n} {a} {A} {l}'),
			'outputs': (    '{n} {a} {A}',     '{n} {a} {A} {l}')
		}[src]

		def gen_output():
			for e in self.vouts_info[src]:
				mmid = e.twlabel.twmmid if e.twlabel else None
				if not mmid:
					yield fs1.format(
						i = CoinTxID(e.txid).hl(color=color),
						n = (nocolor, red)[color](str(e.data['n']).ljust(3)),
						a = CoinAddr(self.proto, e.coin_addr).fmt(
							addr_view_pref, self.max_addrlen[src], color=color)
								if e.coin_addr != self.no_address_str else
							CoinAddr.fmtc(e.coin_addr, self.max_addrlen[src], color=color),
						A = self.proto.coin_amt(e.data['value']).fmt(color=color)
					).rstrip()
				else:
					bal_star, co = ('*', 'melon') if mmid in self.unspent_info else ('', 'brown')
					addr_out = mmid if mmid.type == 'mmgen' else mmid.split(':', 1)[1]
					yield fs2.format(
						i = CoinTxID(e.txid).hl(color=color),
						n = (nocolor, red)[color](str(e.data['n']).ljust(3)),
						a = TwMMGenID.hl2(
							TwMMGenID,
							s = '{:{w}}'.format(addr_out + bal_star, w=self.max_addrlen[src]),
							color = color,
							color_override = co),
						A = self.proto.coin_amt(e.data['value']).fmt(color=color),
						l = e.twlabel.comment.hl(color=color)
					).rstrip()

		return f'\n{indent}'.join(gen_output()).strip()

	def vouts_disp(self, src, width, color, addr_view_pref):

		def gen_output():

			nonlocal space_left

			for e in self.vouts_info[src]:
				mmid = e.twlabel.twmmid if e.twlabel else None
				bal_star, addr_w, co = ('*', 16, 'melon') if mmid in self.unspent_info else ('', 15, 'brown')
				if not mmid:
					if width and space_left < addr_w:
						break
					yield (
						CoinAddr(self.proto, e.coin_addr).fmt(addr_view_pref, addr_w, color=color)
							if e.coin_addr != self.no_address_str else
						CoinAddr.fmtc(e.coin_addr, addr_w, color=color))
					space_left -= addr_w
				elif mmid.type == 'mmgen':
					mmid_disp = mmid + bal_star
					if width and space_left < len(mmid_disp):
						break
					yield TwMMGenID.hl2(TwMMGenID, s=mmid_disp, color=color, color_override=co)
					space_left -= len(mmid_disp)
				else:
					if width and space_left < addr_w:
						break
					yield TwMMGenID.hl2(
						TwMMGenID,
						s = CoinAddr.fmtc(mmid.split(':', 1)[1] + bal_star, addr_w),
						color = color,
						color_override = co)
					space_left -= addr_w
				space_left -= 1

		space_left = width or 0

		return ' '.join(gen_output()) + ' ' * (space_left + 1 if width else 0)

	def amt_disp(self, show_total_amt):
		return (
			self.outputs_total if show_total_amt else
			self.wallet_outputs_total)

	def fee_disp(self, color):
		atomic_unit = self.proto.coin_amt.units[0]
		return '{} {}'.format(
			self.fee.hl(color=color),
			(nocolor, pink)[color]('({:,} {}s/byte)'.format(
				self.fee.to_unit(atomic_unit) // self.vsize,
				atomic_unit)))

class BitcoinTwTxHistory(BitcoinTwView, TwTxHistory, BitcoinTwRPC):

	has_age = True
	hdr_lbl = 'transaction history'
	desc = 'transaction history'
	item_desc = 'transaction'
	item_desc_pl = 'transactions'
	prompt_fs_in = [
		'Sort options: [t]xid, [a]mt, total a[m]t, [A]ge, block[n]um, [r]everse',
		'Column options: toggle [D]ays/date/confs/block, tx[i]d, [T]otal amt',
		'View/Print: pager [v]iew, full pager [V]iew, [p]rint, full [P]rint{s}',
		'Filters/Actions: show [u]nconfirmed, [q]uit menu, r[e]draw:']
	prompt_fs_repl = {
		'BCH': (1, 'Column options: toggle [D]ate/confs, cas[h]addr, tx[i]d, [T]otal amt')
	}
	key_mappings = {
		'A':'s_age',
		'n':'s_blockheight',
		'a':'s_amt',
		'm':'s_total_amt',
		't':'s_txid',
		'r':'s_reverse',
		'D':'d_days',
		'e':'d_redraw',
		'u':'d_show_unconfirmed',
		'i':'d_show_txid',
		'T':'d_show_total_amt',
		'v':'a_view',
		'V':'a_view_detail',
		'p':'a_print_squeezed',
		'P':'a_print_detail'}

	async def get_rpc_data(self):
		blockhash = (
			await self.rpc.call('getblockhash', self.sinceblock)
				if self.sinceblock else '')
		# bitcoin-cli help listsinceblock:
		# Arguments:
		# 1. blockhash            (string, optional) If set, the block hash to list transactions since,
		#                         otherwise list all transactions.
		# 2. target_confirmations (numeric, optional, default=1) Return the nth block hash from the main
		#                         chain. e.g. 1 would mean the best block hash. Note: this is not used
		#                         as a filter, but only affects [lastblock] in the return value
		# 3. include_watchonly    (boolean, optional, default=true for watch-only wallets, otherwise
		#                         false) Include transactions to watch-only addresses
		# 4. include_removed      (boolean, optional, default=true) Show transactions that were removed
		#                         due to a reorg in the "removed" array (not guaranteed to work on
		#                         pruned nodes)
		return (await self.rpc.call('listsinceblock', blockhash, 1, True, False))['transactions']

	async def gen_data(self, rpc_data, lbl_id):

		def gen_parsed_data():
			for o in rpc_data:
				if lbl_id in o:
					l = get_tw_label(self.proto, o[lbl_id])
				else:
					assert o['category'] == 'send', f"{o['address']}: {o['category']} != 'send'"
					l = None
				o.update({
					'twmmid': l.mmid if l else None,
					'comment': (l.comment or '') if l else None,
				})
				yield o

		data = list(gen_parsed_data())

		if self.cfg.debug_tw:
			import json
			from ....rpc.util import json_encoder
			def do_json_dump(*data):
				nw = f'{self.proto.coin.lower()}-{self.proto.network}'
				for d, fn_stem in data:
					with open(f'/tmp/{fn_stem}-{nw}.json', 'w') as fh:
						fh.write(json.dumps(d, cls=json_encoder))

		_mmp = namedtuple('mmap_datum', ['twmmid', 'comment'])

		mm_map = {
			i['address']: (
				_mmp(TwMMGenID(self.proto, i['twmmid']), TwComment(i['comment']))
					if i['twmmid'] else _mmp(None, None)
			)
			for i in data if 'address' in i}

		if self.sinceblock: # mapping data may be incomplete for inputs, so update from 'listlabels'
			mm_map.update(
				{e.coinaddr: _mmp(e.label.mmid, e.label.comment) if e.label else _mmp(None, None)
					for e in await self.get_label_addr_pairs()}
			)

		msg_r('Getting wallet transactions...')
		_wallet_txs = await self.rpc.gathered_icall(
			'gettransaction',
			[(i, True, True) for i in {d['txid'] for d in data}])
		msg('done')

		if not 'decoded' in _wallet_txs[0]:
			_decoded_txs = iter(
				await self.rpc.gathered_call(
					'decoderawtransaction',
					[(d['hex'],) for d in _wallet_txs]))
			for tx in _wallet_txs:
				tx['decoded'] = next(_decoded_txs)

		if self.cfg.debug_tw:
			do_json_dump((_wallet_txs, 'wallet-txs'),)

		_wip = namedtuple('prevout', ['txid', 'vout'])
		txdata = [
			{
				'tx': tx,
				'wallet_vouts': sorted({i.vout for i in
					[_wip(CoinTxID(d['txid']), d['vout']) for d in data]
						if i.txid == tx['txid']}),
				'prevouts': [_wip(CoinTxID(vin['txid']), vin['vout']) for vin in tx['decoded']['vin']]
			}
				for tx in _wallet_txs]

		_prevout_txids = {i.txid for d in txdata for i in d['prevouts']}

		msg_r('Getting input transactions...')
		_prevout_txs = await self.rpc.gathered_call('getrawtransaction', [(i, True) for i in _prevout_txids])
		msg('done')

		_prevout_txs_dict = dict(zip(_prevout_txids, _prevout_txs))

		for d in txdata:
			d['prevout_txs'] = [_prevout_txs_dict[txid] for txid in {i.txid for i in d['prevouts']}]

		if self.cfg.debug_tw:
			do_json_dump(
				(rpc_data,     'txhist-rpc'),
				(data,         'txhist'),
				(mm_map,       'mmap'),
				(_prevout_txs, 'prevout-txs'),
				(txdata,       'txdata'),
			)

		unspent_info = await self.get_unspent_by_mmid()

		return (
			BitcoinTwTransaction(
				parent       = self,
				proto        = self.proto,
				rpc          = self.rpc,
				idx          = idx,
				unspent_info = unspent_info,
				mm_map       = mm_map,
				**d) for idx, d in enumerate(txdata))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.incog_hex: hexadecimal incognito wallet class
"""

from ..util2 import pretty_hexdump, decode_pretty_hexdump
from .incog_base import wallet

class wallet(wallet):

	desc = 'hex incognito data'
	file_mode = 'text'
	no_tty = False

	def _deformat(self):
		ret = decode_pretty_hexdump(self.fmt_data)
		if ret:
			self.fmt_data = ret
			return super()._deformat()
		else:
			return False

	def _format(self):
		super()._format()
		self.fmt_data = pretty_hexdump(self.fmt_data)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.rpc: Tracking wallet RPC class for the MMGen suite
"""

from ..objmethods import MMGenObject

class TwRPC:

	def __new__(cls, proto, *args, **kwargs):
		return MMGenObject.__new__(proto.base_proto_subclass(cls, 'tw.rpc'))

	def __init__(self, proto, rpc, twctl):
		self.proto = proto
		self.rpc = rpc
		self.twctl = twctl



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.bch.params: Bitcoin Cash protocol
"""

from ...protocol import decoded_addr_multiview
from ...addr import CoinAddr
from ..btc.params import mainnet, _finfo
from ..btc.common import b58chk_decode, b58chk_encode
from .cashaddr import cashaddr_decode_addr, cashaddr_encode_addr, cashaddr_addr_types

class mainnet(mainnet):
	is_fork_of      = 'Bitcoin'
	mmtypes         = ('L', 'C')
	preferred_mmtypes = ('C',)
	sighash_type    = 'ALL|FORKID'
	forks = [
		_finfo(478559, '000000000000000000651ef99cb9fcbe0dadde1d424bd9f15ff20136191a5eec', 'BTC', False)
	]
	caps = ()
	coin_amt        = 'BCHAmt'
	max_tx_fee      = 0.1
	cashaddr_pfx    = 'bitcoincash'
	cashaddr        = True

	def decode_addr(self, addr):
		if len(addr) >= 42: # cashaddr
			if addr.islower():
				pass
			elif addr.isupper():
				addr = addr.lower()
			else:
				raise ValueError(f'{addr}: address has mixed case!')
			if ':' in addr:
				assert addr.startswith(self.cashaddr_pfx), f'{addr}: address has invalid prefix!'
			else:
				addr = f'{self.cashaddr_pfx}:{addr}'
			dec = cashaddr_decode_addr(addr)
			ver_bytes = self.addr_fmt_to_ver_bytes[dec.addr_type]
			return decoded_addr_multiview(
				dec.bytes,
				ver_bytes,
				dec.addr_type,
				addr,
				[dec.payload, b58chk_encode(ver_bytes+dec.bytes)] if len(dec.bytes) == self.addr_len else
				[dec.payload],
				0)
		else:
			dec = self.decode_addr_bytes(b58chk_decode(addr))
			enc = cashaddr_encode_addr(
				cashaddr_addr_types[dec.fmt],
				len(dec.bytes),
				self.cashaddr_pfx,
				dec.bytes)
			return decoded_addr_multiview(*dec, enc.addr, [enc.payload, addr], 1)

	def pubhash2addr(self, pubhash, addr_type):
		return CoinAddr(
			self,
			cashaddr_encode_addr(
				cashaddr_addr_types[addr_type],
				len(pubhash),
				self.cashaddr_pfx,
				pubhash).addr
				if self.cfg.cashaddr else
			b58chk_encode(self.addr_fmt_to_ver_bytes[addr_type] + pubhash))

	def pubhash2redeem_script(self, pubhash):
		raise NotImplementedError

	def pubhash2segwitaddr(self, pubhash):
		raise NotImplementedError

class testnet(mainnet):
	addr_ver_info  = {'6f': 'p2pkh', 'c4': 'p2sh'}
	wif_ver_num    = {'std': 'ef'}
	cashaddr_pfx   = 'bchtest'

class regtest(testnet):
	halving_interval = 150
	cashaddr_pfx     = 'bchreg'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.vm.tx.info: transaction info methods for VM chains
"""

from ....util import fmt
from ....addr import MMGenID

def mmid_disp(io, nonmm_str):
	return ' ' + (io.mmid.hl() if io.mmid else MMGenID.hlc(nonmm_str))

class VmTxInfo:

	txinfo_hdr_fs = '{hdr}\n  ID={i} ({a} {c}) Sig={s}\n'
	txinfo_hdr_fs_short = 'TX {i} ({a} {c}) Sig={s}\n'
	txinfo_ftr_fs = fmt("""
		Total in account:  {i} {d}
		Total to spend:    {o} {d}
		Remaining balance: {C} {d}
		TX fee:            {a} {c}{r}
	""")



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.signed: Bitcoin signed transaction class
"""

from ....tx import signed as TxBase
from ....util import fmt, die
from .completed import Completed

class Signed(Completed, TxBase.Signed):

	def compare_size_and_estimated_size(self, tx_decoded):
		est_vsize = self.estimate_size()
		d = tx_decoded
		vsize = d['vsize'] if 'vsize' in d else d['size']
		self.cfg._util.vmsg(f'\nVsize: {vsize} (true) {est_vsize} (estimated)')
		ratio = float(est_vsize) / vsize
		if not (0.95 < ratio < 1.05): # allow for 5% error
			die('BadTxSizeEstimate', fmt(f"""
				Estimated transaction vsize is {ratio:1.2f} times the true vsize
				Your transaction fee estimates will be inaccurate
				Please re-create and re-sign the transaction using the option --vsize-adj={1/ratio:1.2f}
			""").strip())

class AutomountSigned(TxBase.AutomountSigned, Signed):
	pass



#!/usr/bin/env python3

"""
test.modtest_d.misc: miscellaneous unit tests for the MMGen suite
"""

import re, time
from collections import namedtuple

from mmgen.color import yellow, blue, brown
from ..include.common import vmsg

class unit_tests:

	def format_elapsed_hr(self, name, ut, desc='function util.format_elapsed_hr()'):
		from mmgen.util2 import format_elapsed_hr

		vectors = (
			# t now               rel_now show_secs  out
			(1, 1,                     False, False, '0 minutes'),
			(1, 1,                     True,  False, 'just now'),
			(0, -2,                    True,  False, 'just now'),
			(0, -62,                   True,  False, '1 minute in the future'),
			(0, -62,                   False, False, '1 minute (negative elapsed)'),
			(0, 7,                     False, False, '0 minutes'),
			(0, 7,                     True,  False, 'just now'),
			(0, 60,                    True,  False, '1 minute ago'),
			(0, 67,                    True,  False, '1 minute ago'),
			(0, 7200,                  True,  False, '2 hours ago'),
			(0, 3600 + 180 + 1,        True,  False, '1 hour 3 minutes ago'),
			(0, 3600*27 + 180 + 7,     True,  False, '1 day 3 hours 3 minutes ago'),
			(0, 3600*24*367 + 180 + 7, True,  False, '367 days 3 minutes ago'),

			(1, 1,                     False,  True, '0 seconds'),
			(1, 1,                     True,   True, 'just now'),
			(0, -2,                    True,   True, '2 seconds in the future'),
			(0, -62,                   True,   True, '1 minute 2 seconds in the future'),
			(0, -62,                   False,  True, '1 minute 2 seconds (negative elapsed)'),
			(0, 7,                     False,  True, '7 seconds'),
			(0, 7,                     True,   True, '7 seconds ago'),
			(0, 60,                    False,  True, '1 minute'),
			(0, 67,                    False,  True, '1 minute 7 seconds'),
			(0, 7200,                  False,  True, '2 hours'),
			(0, 3600 + 180 + 1,        False,  True, '1 hour 3 minutes 1 second'),
			(0, 3600*27 + 180 + 7,     False,  True, '1 day 3 hours 3 minutes 7 seconds'),
			(0, 3600*24*367 + 180 + 7, True,   True, '367 days 3 minutes 7 seconds ago'),
		)

		fs = '    {:7}  {:9}  {:<8}    {}'
		vmsg(brown('  vectors:'))
		vmsg(fs.format('REL_NOW', 'SHOW_SECS', 'ELAPSED', 'OUTPUT'))
		for (t, now, rel_now, show_secs, out_chk) in vectors:
			out = format_elapsed_hr(t, now=now, rel_now=rel_now, show_secs=show_secs)
			assert out == out_chk, f'{out} != {out_chk}'
			vmsg(fs.format(repr(rel_now), repr(show_secs), now-t, out))

		vmsg(brown('  real time:'))
		start = time.time() - 3600 - 127
		ret = format_elapsed_hr(start) # test old default behavior
		vmsg(f'    {3600 - 127:<8}    {ret}')
		ret = format_elapsed_hr(start, show_secs=True)
		vmsg(f'    {3600 - 127:<8}    {ret}')

		return True

	def xmrwallet_uarg_info(self, name, ut, desc='dict xmrwallet.xmrwallet_uarg_info'): # WIP
		from mmgen.xmrwallet import uarg_info
		vs = namedtuple('vector_data', ['text', 'groups'])
		fs = '{:16} {}'

		vmsg(blue('  ' + fs.format('ID', 'ANNOT')))
		for k, v in uarg_info.items():
			vmsg('  ' + fs.format(k, v[0]))

		vectors = {
			'sweep_spec': (
				vs('1:2',     "('1', '2', None, None)"),
				vs('1:2,3',   "('1', '2', '3', None)"),
				vs('1:2,3:4', "('1', '2', '3', '4')"),
			),
		}

		vmsg('')
		for k, v in uarg_info.items():
			vmsg(f'  {k}')
			if k in vectors:
				vmsg(f'    pat: {v.pat}')
				vmsg( '    vectors:')
				for vec in vectors[k]:
					m = re.match(v.pat, vec.text)
					vmsg(f'      {vec.text:10} ==> {m.groups()}')
					assert str(m.groups()) == vec.groups
			else:
				vmsg(yellow('    TBD'))

		return True

	def pyversion(self, name, ut, desc='class pyversion.PythonVersion'):
		from mmgen.pyversion import PythonVersion, python_version

		ver = {}
		fs = '{:<7} {:<9} {:<5} {}'
		vmsg('\n' + fs.format('Version', 'PyVersion', 'Major', 'Minor'))

		for k in ('current', '3.3', '3.12', '4.3', '7.0'):
			obj = python_version if k == 'current' else PythonVersion(k)
			major, minor = [int(s) for s in obj.split('.')]
			assert obj.major == major and obj.minor == minor
			vmsg(fs.format(k.upper(), obj, major, minor))
			ver[k] = obj

		vmsg('\nPerforming comparison tests:')

		assert ver['3.3'] == '3.3'

		assert ver['current'] < ver['7.0']
		assert ver['3.3']     < ver['4.3']
		assert ver['3.12']    < ver['7.0']
		assert ver['3.3']     < ver['3.12']

		assert ver['current'] < '7.0'
		assert ver['3.3']     < '4.3'
		assert ver['3.12']    < '7.0'
		assert ver['3.3']     < '3.12' # ensure we’re comparing numerically, not stringwise

		assert ver['current'] <= ver['current']
		assert ver['3.3']     <= '4.3'
		assert ver['3.12']    <= '7.0'
		assert ver['3.3']     <= '3.12'

		assert ver['current'] == ver['current']
		assert ver['3.3']     == ver['3.3']
		assert ver['3.3']     != ver['3.12']
		assert ver['3.3']     != ver['4.3']
		assert ver['3.3']     == '3.3'
		assert ver['3.3']     != '3.12'
		assert ver['3.3']     != '4.3'

		assert ver['current'] > '3.6'
		assert ver['7.0']     > ver['current']
		assert ver['4.3']     > '3.3'
		assert ver['3.12']    > '3.3'

		assert ver['current'] >= ver['current']
		assert ver['7.0']     >= ver['current']
		assert ver['4.3']     >= '3.3'
		assert ver['3.12']    >= '3.12'
		assert ver['3.12']    >= '3.3'

		assert '3.0' < ver['3.12'] < '3.13'
		assert '3.9' < ver['3.12'] # ensure we’re reverse comparing numerically, not stringwise
		assert '3.3' < ver['4.3']  <= '4.3'
		assert '4.3' <= ver['4.3'] <= '4.3'
		assert '4.3' == ver['4.3'] == '4.3'

		return True

	def exp_int(self, name, ut, desc='ExpInt() class'):
		from mmgen.util2 import ExpInt

		for num, trunc, chk, out in (
				('0e0',              0,                  0,                  '0'),
				('0e1',              0,                  0,                  '0'),
				('3e0',              3,                  3,                  '3'),
				('3e1',              30,                 30,                 '30'),
				('3e2',              300,                300,                '300'),
				('3e3',              3000,               3000,               '3e3'),
				(3000,               3000,               3000,               '3e3'),
				(1,                  1,                  1,                  '1'),
				('123e0',            123,                123,                '123'),
				(123,                123,                123,                '123'),
				(12345,              12340,              12345,              '12345'),
				(123456,             123400,             123456,             '123456'),
				(1234567,            1234000,            1234000,            '1234e3'),
				(123456789,          123400000,          123400000,          '1234e5'),
				(998877665544332211, 998800000000000000, 998800000000000000, '9988e14'),
			):
			e = ExpInt(num, prec=4)
			enc = e.enc
			assert enc == out, f'{enc} != {out}'
			assert e.trunc == trunc, f'{e.trunc} != {trunc}'
			vmsg('')
			vmsg(f'num        {num}')
			vmsg(f'enc        {enc}')
			dec = ExpInt(enc, prec=4)
			vmsg(f'enc -> dec {dec}')
			vmsg(f'chk        {chk}')
			assert dec == chk, f'{dec} != {chk}'
			dec_enc = dec.enc
			vmsg(f'dec -> enc {dec_enc}')
			vmsg(f'out        {out}')
			assert dec_enc == out, f'{dec_enc} != {out}'

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.store: Tracking wallet control class with store
"""

import json
from pathlib import Path

from ..base_obj import AsyncInit
from ..obj import TwComment
from ..util import msg, die, cached_property
from ..addr import is_coin_addr, is_mmgen_id, CoinAddr

from .shared import TwMMGenID, TwLabel
from .ctl import TwCtl, write_mode, label_addr_pair

class TwCtlWithStore(TwCtl, metaclass=AsyncInit):

	caps = ('batch',)
	tw_fn = 'tracking-wallet.json'
	aggressive_sync = False

	async def __init__(
			self,
			cfg,
			proto,
			*,
			mode              = 'r',
			token_addr        = None,
			no_rpc            = False,
			no_wallet_init    = False,
			rpc_ignore_wallet = False):

		await super().__init__(cfg, proto, mode=mode, no_rpc=no_rpc, rpc_ignore_wallet=rpc_ignore_wallet)

		self.cur_balances = {} # cache balances to prevent repeated lookups per program invocation

		if cfg.cached_balances:
			self.use_cached_balances = True

		self.tw_dir = Path(
			self.cfg.data_dir_root,
			'altcoins',
			self.proto.coin.lower(),
			('' if self.proto.network == 'mainnet' else self.proto.network)
		)
		self.tw_path = self.tw_dir / self.tw_fn

		if no_wallet_init:
			return

		self.init_from_wallet_file()

		if self.data['coin'] != self.proto.coin:
			fs = 'Tracking wallet coin ({}) does not match current coin ({})!'
			die('WalletFileError', fs.format(self.data['coin'], self.proto.coin))

		self.conv_types(self.data[self.data_key])

	def __del__(self):
		"""
		TwCtl instances opened in write or import mode must be explicitly destroyed with ‘del
		twuo.twctl’ and the like to ensure the instance is deleted and wallet is written before
		global vars are destroyed by the interpreter at shutdown.

		Not that this code can only be debugged by examining the program output, as exceptions
		are ignored within __del__():

			/usr/share/doc/python3.6-doc/html/reference/datamodel.html#object.__del__

		Since no exceptions are raised, errors will not be caught by the test suite.
		"""
		if getattr(self, 'mode', None) == 'w': # mode attr might not exist in this state
			self.write()
		elif self.cfg.debug:
			msg('read-only wallet, doing nothing')

	def upgrade_wallet_maybe(self):
		pass

	def conv_types(self, ad):
		for k, v in ad.items():
			if k not in ('params', 'coin'):
				v['mmid'] = TwMMGenID(self.proto, v['mmid'])
				v['comment'] = TwComment(v['comment'])

	def init_empty(self):
		self.data = {
			'coin': self.proto.coin,
			'network': self.proto.network.upper(),
			'addresses': {},
		}

	def init_from_wallet_file(self):
		from ..fileutil import check_or_create_dir, get_data_from_file
		check_or_create_dir(self.tw_dir)
		try:
			self.orig_data = get_data_from_file(self.cfg, self.tw_path, quiet=True)
			self.data = json.loads(self.orig_data)
		except:
			try:
				self.tw_path.stat()
			except:
				self.orig_data = ''
				self.init_empty()
				self.force_write()
			else:
				die('WalletFileError', f'File ‘{self.tw_path}’ exists but does not contain valid JSON data')
		else:
			self.upgrade_wallet_maybe()

		# ensure that wallet file is written when user exits via KeyboardInterrupt:
		if self.mode == 'w':
			import atexit
			def del_twctl(twctl):
				self.cfg._util.dmsg(f'Running exit handler del_twctl() for {twctl!r}')
				del twctl
			atexit.register(del_twctl, self)

	@write_mode
	async def batch_import_address(self, args_list):
		return [await self.import_address(a, label=b, rescan=c) for a, b, c in args_list]

	async def rescan_addresses(self, coin_addrs):
		pass

	@write_mode
	async def import_address(self, addr, *, label, rescan=False):
		r = self.data_root
		if addr in r:
			if self.check_import_mmid(addr, r[addr]['mmid'], label.mmid):
				r[addr]['mmid'] = label.mmid
			if label.comment: # overwrite existing comment only if new comment not empty
				r[addr]['comment'] = label.comment
		else:
			r[addr] = {'mmid': label.mmid, 'comment': label.comment}

	@write_mode
	async def remove_address(self, addr):
		r = self.data_root

		if is_coin_addr(self.proto, addr):
			have_match = lambda k: k == addr
		elif is_mmgen_id(self.proto, addr):
			have_match = lambda k: r[k]['mmid'] == addr
		else:
			die(1, f'{addr!r} is not an Ethereum address or MMGen ID')

		for k in r:
			if have_match(k):
				# return the addr resolved to mmid if possible
				ret = r[k]['mmid'] if is_mmgen_id(self.proto, r[k]['mmid']) else addr
				del r[k]
				self.write()
				return ret
		msg(f'Address {addr!r} not found in {self.data_root_desc!r} section of tracking wallet')
		return None

	@write_mode
	async def set_label(self, coinaddr, lbl):
		for addr, d in list(self.data_root.items()):
			if addr == coinaddr:
				d['comment'] = lbl.comment
				self.write()
				return True
		msg(f'Address {coinaddr!r} not found in {self.data_root_desc!r} section of tracking wallet')
		return False

	@property
	def sorted_list(self):
		return sorted([{
				'addr':    x[0],
				'mmid':    x[1]['mmid'],
				'comment': x[1]['comment']
			} for x in self.data_root.items() if x[0] not in ('params', 'coin')],
			key = lambda x: x['mmid'].sort_key + x['addr'])

	@property
	def mmid_ordered_dict(self):
		return dict((x['mmid'], {'addr': x['addr'], 'comment': x['comment']}) for x in self.sorted_list)

	async def get_label_addr_pairs(self):
		return [label_addr_pair(
				TwLabel(self.proto, f"{mmid} {d['comment']}"),
				CoinAddr(self.proto, d['addr'])
			) for mmid, d in self.mmid_ordered_dict.items()]

	@cached_property
	def used_addrs(self):
		from decimal import Decimal
		# TODO: for now, consider used addrs to be addrs with balance
		return ({k for k, v in self.data['addresses'].items() if Decimal(v.get('balance', 0))})

	@property
	def data_root(self):
		return self.data[self.data_key]

	@property
	def data_root_desc(self):
		return self.data_key

	def cache_balance(self, addr, bal, *, session_cache, data_root, force=False):
		if force or addr not in session_cache:
			session_cache[addr] = str(bal)
			if addr in data_root:
				data_root[addr]['balance'] = str(bal)
				if self.aggressive_sync:
					self.write()

	def get_cached_balance(self, addr, session_cache, data_root):
		if addr in session_cache:
			return self.proto.coin_amt(session_cache[addr])
		if self.use_cached_balances:
			return self.proto.coin_amt(
				data_root[addr]['balance'] if addr in data_root and 'balance' in data_root[addr]
				else '0')

	async def get_balance(self, addr, *, force_rpc=False, block='latest'):
		ret = None if force_rpc else self.get_cached_balance(addr, self.cur_balances, self.data_root)
		if ret is None:
			ret = await self.rpc_get_balance(addr, block=block)
			if ret is not None:
				self.cache_balance(addr, ret, session_cache=self.cur_balances, data_root=self.data_root)
		return ret

	def force_write(self):
		mode_save = self.mode
		self.mode = 'w'
		self.write()
		self.mode = mode_save

	@write_mode
	def write_changed(self, data, quiet):
		from ..fileutil import write_data_to_file
		write_data_to_file(
			self.cfg,
			self.tw_path,
			data,
			desc              = f'{self.base_desc} data',
			ask_overwrite     = False,
			ignore_opt_outdir = True,
			quiet             = quiet,
			check_data        = True, # die if wallet has been altered by another program
			cmp_data          = self.orig_data)

		self.orig_data = data

	def write(self, *, quiet=True):
		self.cfg._util.dmsg(f'write(): checking if {self.desc} data has changed')

		wdata = json.dumps(self.data)
		if self.orig_data != wdata:
			self.write_changed(wdata, quiet=quiet)
		elif self.cfg.debug:
			msg('Data is unchanged\n')
